<?php

/**
 * Plugin Name:     WordPress Support
 * Plugin URI:      http://wp-support.co/
 * Description:     NOTE: This plugin is currently in BETA development and will be released soon!
 * Version:         0.8.3
 * Author:          Q Studio
 * Author URI:      http://qstudio.us
 * License:         GPL2
 * Class:           Q_Support
 * Extends          Q_Plugin
 * Text Domain:     q-support
 */

/*  

    Copyright 2014 Q Studio ( url : http://qstudio.us )

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as 
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

// quick check :) ##
defined( 'ABSPATH' ) OR exit;

/* Plugin Class */
if ( ! class_exists( 'Q_Support' ) ) 
{
    
    // Q_Plugin library ##
    require_once plugin_dir_path( __FILE__ ) . 'library/Q_Plugin.class.php';
        
    // Poll ##
    require_once plugin_dir_path( __FILE__ ) . 'library/q_support_poll.php';
    
    // plugin version
    define( 'Q_SUPPORT_VERSION', '0.8.3' ); // version ##

    // cache ##
    define( 'Q_SUPPORT_CACHE', true ); // cache and transients ##
    define( 'Q_SUPPORT_CACHE_EXPIRE', 60*60*24*7 ); // cache timeout ##

    // error logs ##
    define( 'Q_SUPPORT_LOG_FILE',  plugin_dir_path( __FILE__ )."logs/debug.log" ); // path to file ##
    define( 'Q_SUPPORT_LOG_PATH',  plugin_dir_path( __FILE__ )."logs/" ); // path to log folder ##
    define( 'Q_SUPPORT_LOG_MAX_KB', 1024 ); // max log file size - 1 MB for now ##

    // date ##
    define( 'Q_SUPPORT_DATE_FORMAT', 'Y-m-d H:i:s' );

    // data whitelisting ##
    define( 'Q_SUPPORT_WHITELIST_DOMAIN', 'domain.com' ); // no domains ##

    // third party domain ##
    define( 'Q_SUPPORT_DOMAIN', 'http://www.wp-support.co/' ); // for links out ##

    // required capability to fully use this plugin ##
    define( 'Q_SUPPORT_CAPABILITY', 'edit_posts' ); // to view settings and make support requests ##
    define( 'Q_SUPPORT_CAPABILITY_ADMIN', 'manage_options' ); // to debug and view error.log ##
    
    // on plugin uninstall ##
    register_uninstall_hook( __FILE__, array( 'Q_Support',  'uninstall_hook' ) );
    
    // declare the class and base Class ##
    class Q_Support 
    {
        
        // settings API ##
        #public $q_support_settings;
        public $plugin_url = 'edit.php?post_type=q_support';
        #public $plugin_url_connect = 'edit.php?post_type=q_support&page=connect';
        public $plugin_url_settings = 'edit.php?post_type=q_support&page=settings';
        public $plugin_url_data = 'edit.php?post_type=q_support&page=data';
        public $text_domain = 'q-support'; // for translation ##
        public $q_support_api_url; // API URL ##
        public $q_support_api_timeout = 60; // allows us to extend the timeout on API calls ##
        
        // default q_support_settings ##
        public $q_support_settings_default = array(
            'target'        => "community"
            ,'agent'        => "0"
            ,'wordpress'    => "no"
            ,'server'       => "no"
            ,'client'       => "no"
        );
        
        // to clean up links ##
        public $allowed_tags_href = array(
            'a' => array(
                'href' => array()
                ,'title' => array()
            )
            ,'abbr' => array(
                'title' => array()
            )
            ,'acronym' => array(
                'title' => array()
            )
            ,'code' => array()
            ,'em' => array()
            ,'strong' => array()
        );
            
        // to clean up the desc ##
        public $allowed_tags_desc = array(
            'abbr' => array(
                'title' => array()
            )
            ,'acronym' => array(
                'title' => array()
            )
            ,'code' => array()
            ,'em' => array()
            ,'strong' => array()
        );
        
        // help_tab variables ##
        public $help_tabs = array();
        
        // question categories ##
        public $question_categories = array ();

        
        /**
         * 
         * Instatiate Class
         * 
         * @since       0.2
         * @return      void
         */
        public function __construct() 
        {
            
            // on plugin activation ##
            register_activation_hook( __FILE__, array( $this, 'activation_hook' ) );

            // on plugin deactivation ##
            register_deactivation_hook( __FILE__, array( $this, 'deactivation_hook' ) );

            // error logging ##
            add_action( 'admin_init', array( $this, 'set_error_logging' ), 1 );
            
            // front-end options ##
            if ( ! is_admin() ) {
                
                // add link to request help from admin bar ##
                add_action( 'admin_bar_menu', array ( $this, 'admin_bar_menu' ), 100 );
                
                // add scripts ##
                add_action( 'wp_enqueue_scripts', array( $this, 'wp_enqueue_scripts' ), 100000 );
                
            } else { // admin features ##
                
                // error logging ##
                add_action( 'admin_init', array( $this, 'set_error_logging' ), 1 );
                
                // add link to request help from admin bar ##
                add_action( 'admin_bar_menu', array ( $this, 'admin_bar_menu' ), 100 );
                
                // admin_init ##
                add_action('admin_init', array( $this, 'admin_init' ), 2 );

                // add admin menus ##
                add_action( 'admin_menu', array( $this, 'add_admin_menu_settings' ), 1 ); // plugin settings ##
                add_action( 'admin_menu', array( $this, 'add_admin_menu_errors' ), 2 ); // error log viewer ##
                add_action( 'admin_menu', array( $this, 'add_admin_menu_data' ), 3 ); // plugin data viewer ##

                // check for q_support_options ##
                add_action( 'admin_init', array( $this, 'get_q_support_options' ), 3 );

                // add custom post type ##
                add_action( 'init', array( $this, 'register_post_type' ), 1 );

                // modify cpt post row actions ##
                add_action( 'post_row_actions', array( $this, 'post_row_actions' ), 10, 2 );

                // modify cpt columns ###
                add_action( 'admin_init', array( $this, 'manage_columns' ), 4 );
                
                // user columns ##
                add_action( 'manage_users_custom_column', array( $this, 'manage_users_custom_column' ), 0, 3 );
                add_filter( 'manage_users_columns', array( $this,  'manage_users_columns' ) );
                
                // limit results to current user ##
                add_action( 'admin_init', array( $this, 'limit_results' ), 5 );
                
                // upgrade and activation notices ##
                add_action( 'admin_init', array( $this, 'admin_update_notice' ), 5 );
                add_action( 'admin_notices', array( $this, 'admin_notices' ) );

                // add settings link to plugin ##
                add_filter( "plugin_action_links_".plugin_basename(__FILE__), array( $this, 'add_plugin_settings_link' ) );

                // set text domain
                add_action( 'admin_init', array( $this, 'load_plugin_textdomain' ), 6 );

                // add scripts ##
                add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_scripts' ), 100000 );

                // add help tab to admin UI ##
                add_action( 'in_admin_header', array( $this, 'add_help_tab' ), 0 );
                
                // ajax hooks ##
                add_action('wp_ajax_q_support_connect', array( $this, 'ajax_q_support_connect' )); // connect user account to API ##
                add_action('wp_ajax_q_support_settings', array( $this, 'ajax_q_support_settings' )); // clear saved support data ##
                add_action('wp_ajax_q_support_form', array( $this, 'ajax_q_support_form' )); // save support request ##
                add_action('wp_ajax_q_support_data_clear', array( $this, 'ajax_q_support_data_clear' )); // clear saved support data ##
                
                // @todo - if user edits their profile email - we need to alert API of the new email address ##
                
                // clean up when a user deletes a local question ##
                add_action( 'before_delete_post', array( $this, 'before_delete_post' ) );
                
            }
            
        }
        
        
        /**
         * Plugin Activation
         * 
         * @since       0.3
         * @return      void
         */
        public function activation_hook()
        {
            
            // get plugin options ##
            $q_support_options = $this->get_q_support_options();
            
            // if the user who actiated the plugin is not connected to the API - give them a notice to Connect ##
            if ( ! $this->get_q_support_user_connected() ) {
            
                // get plugin name ##
                $name = get_file_data( __FILE__, array ( 'Plugin Name' ), 'plugin' );

                // save notices to option ##
                $notices = array();
                $notices[] = sprintf ( 
                    __(
                        '<strong>%s</strong> - To submit support requests, you need to <a href="#" class="button q_support_connect q_support_space_both">%s</a> your account - <a href="%s">%s</a>',
                        $this->text_domain
                    )
                    ,esc_html( __("$name[0] Activated", $this->text_domain ) ) 
                    #,esc_url( $this->plugin_url_connect ) // connect to API ##
                    ,esc_html( __("Connect", $this->text_domain ) ) 
                    ,esc_url( Q_SUPPORT_DOMAIN.'meta/terms-of-use/' ) // terms ##
                    ,esc_html( __("Terms of Use", $this->text_domain ) )
                );

                $q_support_options["deferred_admin_notices"] = $notices;

            }
            
            // add custom post type ##
            $this->register_post_type();
            
            // save plugin version ##
            $q_support_options["version"] = Q_SUPPORT_VERSION; 
            
            // add flag that plugin is configured ##
            $q_support_options["configured"] = true;
            
            // save it all ##
            update_option( 'q_support_options', $q_support_options );
            
            // get and save default settings ##
            $this->get_q_support_settings();
            
            // Redirect if not activating multiple plugins at once
            if( ! isset( $_GET['activate-multi'] ) ) {
                wp_redirect( admin_url( $this->plugin_url ) );
            }
            
        }
        
                
        /**
         * Plugin Deactivation
         * 
         * @since       0.3
         * @return      void
         */
        public function deactivation_hook()
        {
            
            // clear all transient data ##
            $this->delete_cache();
            
            // delete the error log file ##
            $this->delete_error_log();
            
            // array of plugin options ##
            $options = array (
                    'q_support_options' // plugin settings - ok to delete on deactivate ##
                ,   'q_store'           // debugging tracker ##
            );
            
            // remove plugin options ##
            $this->delete_options( $options );
            
            // array of plugin user_meta to delete for all users ##
            $user_meta_keys = array (
                    'q_support_settings'
            );
            
            // batch delete user_meta ##
            $this->delete_user_meta( $user_meta_keys );
            
            // remove cron task ##
            q_support_clear_schedule();
            
        }
        
        
        /**
         * Plugin Uninstalled
         * 
         * @since       0.3
         * @return      void
         */
        public function uninstall_hook()
        {
            
            // array of plugin options ##
            $options = array (
                    'q_support_options'     // plugin settings ##
                ,   'q_store'               // debugging tracker ##
                     
            );
            
            // remove plugin options ##
            $this->delete_options( $options );
            
            // array of plugin user_meta to delete for all users ##
            $user_meta_keys = array (
                    'q_support_settings'
            );
            
            // batch delete user_meta ##
            $this->delete_user_meta( $user_meta_keys );
            
            // clear all transient data ##
            $this->delete_cache();
            
            // delete the error log file ##
            $this->delete_error_log();
            
            // delete the dump directory and all it's contents ##
            $this->delete_directory( plugin_dir_path( __FILE__ )."dump/" );
            
            // delete CPT posts ##
            global $wpdb;

            $query = "
                DELETE FROM wp_posts 
                WHERE post_type = 'q_support' 
                LIMIT 0, 3000
            ";

            $wpdb->query($query);
            
        }
        
        
        /**
         * Delete multiple wp_options keys
         * 
         * @param       array   $options    Array of option names to delete
         * @param       string  $funtion    Function to call to delete options
         * @since       0.3
         * @return      void
         */
        public static function delete_options( $options = null, $function = 'delete_option' ) 
        {
    
            if ( ! $options || ! is_array ( $options ) ) { return; }

            foreach( $options as $option ) {

                call_user_func( $function, $option );

            }

        }
        
        
        /**
         * Delete multiple user_meta keys
         * 
         * @param       array   $keys    Array of option names to delete
         * @since       0.3
         * @return      void
         */
        public static function delete_user_meta( $keys = null ) 
        {
    
            if ( ! $keys || ! is_array ( $keys ) ) { return; }

            foreach( $keys as $key ) {

                $meta_type  = 'user';
                $user_id    = 0; // This will be ignored, since we are deleting for all users.
                $meta_key   = $key;
                $meta_value = ''; // Also ignored. The meta will be deleted regardless of value.
                $delete_all = true;

                delete_metadata( $meta_type, $user_id, $meta_key, $meta_value, $delete_all );

            }

        }
        
        
        /**
         * Admin Init
         * 
         * @since       0.3
         * @return      void
         */
        public function admin_init() 
        {
            
            // get plugin options ##
            $q_support_options = $this->get_q_support_options();
            
            // sort out API URL ##
            $q_support_options["api_url"] = Q_SUPPORT_DOMAIN.'api/';
            
            if ( defined( 'Q_DEBUG' ) && Q_DEBUG === true ) {
                $q_support_options["api_url"] = 'http://localhost/wps/wordpress/api/'; // local testing // @todo - remove on future public release ##
            }
            
            // api url ##
            $this->q_support_api_url = $q_support_options["api_url"];
            
            // api timeout ##
            $q_support_options["api_timeout"] = $this->q_support_api_timeout;
            
            // save options ##
            update_option( 'q_support_options', $q_support_options );
            
            // define question categories ##
            $this->question_categories = array (
                    __('General', $this->text_domain )
                ,   __('Themes', $this->text_domain )
                ,   __('Plugins', $this->text_domain )
                ,   __('Widgets', $this->text_domain )
                ,   __('Advanced', $this->text_domain )
                ,   __('Code', $this->text_domain )
                ,   __('Design', $this->text_domain )
                ,   __('Support', $this->text_domain )
                ,   __('Other', $this->text_domain )
            );            
            
        }
        
        
        /**
          * Add Admin Bar menu item ##
          * 
          * @since      0.3
          * @link       http://blog.rutwick.com/add-items-anywhere-to-the-wp-3-3-admin-bar
          * @return     void
          */
        public function admin_bar_menu( $admin_bar )
        {
            
            if ( ! $this->get_q_support_user_connected() ) { return false; } // not shown until this user is connectd ##
            
            // grab all support requests ##
            $args = array( 
                    'posts_per_page'    => -1  // get all pages
                ,   'post_type'         => 'q_support' 
                ,   'fields'            => 'ids' // Only get post IDs
                ,   'author'            =>  $this->get_current_user("ID")
            );

            $questions = get_posts( $args );
            $questions_string = implode( ',', array_map( 'intval', $questions ) );
            
            // check if there is anything to check ##
            if ( $questions_string ) {
            
                global $wpdb;

                // get total comments ##
                $comments = $wpdb->get_var( $wpdb->prepare(
                    "
                        SELECT SUM(meta_value) AS meta_value_sum
                        FROM {$wpdb->prefix}postmeta 
                        WHERE meta_key = 'q_support_question_comments' AND post_id IN (%s)
                    "
                    ,   $questions_string
                ) );

                // get total answers ##
                $answers = $wpdb->get_var( $wpdb->prepare(
                    "
                    SELECT SUM(meta_value) AS meta_value_sum
                    FROM {$wpdb->prefix}postmeta 
                    WHERE meta_key = 'q_support_question_answers' AND post_id IN (%s)
                    "
                    ,   $questions_string
                ) );
            
            }
                
            // default values and type casting ##
            $comments = isset( $comments ) ? (int)$comments : 0 ;
            $answers = isset ( $answers ) ? (int)$answers : 0 ;
            
            // we need to store, calcuate, update and retrieve data from user meta ##
            $totals = array( 
                    'answers'   =>  $answers
                ,   'comments'  =>  $comments
            );
            
            // save totals to use meta ##
            #$q_support_settings = $this->get_q_support_settings();
            #$q_support_settings["poll_totals"] = $totals;
            #update_user_meta( $this->get_current_user("ID"), 'q_support_settings', $q_support_settings );
            
            // save it all ##
            $this->set_q_support_settings( array( 
                "poll_totals"   => $totals
            ));
            
            $update = sprintf(
                    __(
                        'Support:<span> %s Answers %s Comments</span>'
                        , $this->text_domain 
                    )
                    , "<span class='count count-".$answers."'>{$answers}</span>"
                    , "<span class='count count-".$comments."'>{$comments}</span>"
                );
            
            // The properties of the new item ##
            $args = array(
                'id'    => 'q_support'
                ,'title' => $update
                ,'href'  => admin_url( $this->plugin_url )."&q_clear"
                ,'meta'  => array(
                    'title' => __('Support', $this->text_domain),
                )
                ,'parent' => 'my-account'
            );

            // This is where the magic works. ##
            $admin_bar->add_menu( $args);
            
        }
        
        
        /**
         * Returns all the settings fields
         *
         * @since       0.3
         * @return array settings fields
         */
        function get_q_support_settings_fields() 
        {
            
            return array (
                
                'target'            => array(
                    'name'          => 'target'
                    ,'label'        => __( 'Support Type', $this->text_domain )
                    ,'desc'         => __( 'You can send open support requests to the community, or select one specific agent.', $this->text_domain )
                    ,'type'         => 'select'
                    ,'default'      => 'community'
                    ,'options'      => array(
                        'community' => 'Community'
                        ,'agent'    => 'Dedicated Agent'
                    )
                )

                ,'agent'           => array(
                    'name'          => 'agent'
                    ,'label'        => __( 'Select Agent', $this->text_domain )
                    ,'desc'         => sprintf( 
                                        __( 'Select an agent from the list. <a href="%s" class="_blank">%s</a>.', $this->text_domain )
                                        , esc_url( Q_SUPPORT_DOMAIN.'meta/dedicated-agents/' )
                                        , esc_html( __("Read More", $this->text_domain ) ) 
                                    )
                    ,'type'         => 'select'
                    ,'default'      => '0'
                    ,'options'      => $this->get_agents()
                )

                ,'wordperss'        => array(
                    'name'          => 'wordpress'
                    ,'label'        => __( 'WordPress', $this->text_domain )
                    ,'desc'         => sprintf( 
                                        __( 'Include information about your <a href="%s" class="_blank">%s</a>.', $this->text_domain )
                                        , esc_url( Q_SUPPORT_DOMAIN.'plugin/wordpress-data/' )
                                        , esc_html( __("WordPress Installation", $this->text_domain ) ) 
                                    )
                    ,'type'         => 'select'
                    ,'default'      => 'yes'
                    ,'options'      => array(
                        'yes'       => 'Yes'
                        ,'no'       => 'No'
                    )
                )

                ,'server'           => array(
                    'name'          => 'server'
                    ,'label'        => __( 'Server', $this->text_domain )
                    ,'desc'         => sprintf( 
                                        __( 'Include information about your <a href="%s" class="_blank">%s</a>.', $this->text_domain )
                                        , esc_url( Q_SUPPORT_DOMAIN.'plugin/web-server-data/' )
                                        , esc_html( __("Web Server", $this->text_domain ) ) 
                                    )
                    ,'type'         => 'select'
                    ,'default'      => 'yes'
                    ,'options'      => array(
                        'yes'       => 'Yes'
                        ,'no'       => 'No'
                    )
                )

                ,'client'           => array(
                    'name'          => 'client'
                    ,'label'        => __( 'Client', $this->text_domain )
                    ,'desc'         => sprintf( 
                                        __( 'Include information about your <a href="%s" class="_blank">%s</a>.', $this->text_domain )
                                        , esc_url( Q_SUPPORT_DOMAIN.'plugin/client-data/' )
                                        , esc_html( __("Browser", $this->text_domain ) ) )
                    ,'type'         => 'select'
                    ,'default'      => 'yes'
                    ,'options'      => array(
                        'yes'       => 'Yes'
                        ,'no'       => 'No'
                    )
                )
                    
            );

        }
        
        
        /*
         * Add admin menu
         * 
         * @since       0.3
         * @return      void
         **/
        public function add_admin_menu_settings() 
        {
            
            if ( ! $this->get_q_support_user_connected() ) { return false; } // not shown until this user is connectd ##
            
            // add a sub menu to the CPT ##
            $add_admin_menu_settings = add_submenu_page (
                $this->plugin_url
                ,__( 'Settings', $this->text_domain ) ## page title
                ,__( 'Settings', $this->text_domain ) ## menu title
                ,Q_SUPPORT_CAPABILITY ## capability needed
                ,'settings'
                ,array( $this, 'view_settings' )
            );
            
	}
        
        
        /*
         * Add Error Log View to admin menu
         * 
         * @since       0.3
         * @return      void
         **/
        public function add_admin_menu_errors() 
        {
            
            if ( ! $this->get_q_support_user_connected() || defined( 'WP_DEBUG' ) && WP_DEBUG === true ) { return false; } // not until the plugin is connectd ##
            
            // add a sub menu to the CPT ##
            $add_admin_menu_errors = add_submenu_page (
                $this->plugin_url
                ,__('Error Log', $this->text_domain ) ## page title*/
                ,__('Errors', $this->text_domain ) ## menu title*/
                ,Q_SUPPORT_CAPABILITY ## capability needed
                ,'errors'
                ,array( $this, 'view_error_log' )
            );
            
	}
        
        
        /*
         * Add plugin configuration debugger to admin menus
         * 
         * @since       0.3
         * @return      void
         **/
        public function add_admin_menu_data() 
        {
            
            if ( ! $this->get_q_support_user_connected() ) { return false; } // not until the plugin is connectd ##
            
            // add a sub menu to the CPT ##
            $add_admin_menu_data = add_submenu_page (
                $this->plugin_url
                ,__( 'Plugin Data', $this->text_domain ) ## page title*/
                ,__( 'Data', $this->text_domain ) ## menu title*/
                ,Q_SUPPORT_CAPABILITY ## capability needed
                ,'data'
                ,array( $this, 'view_q_support_data' )
            );
            
	}

        
        /**
         * Plugin Settings Page 
         * 
         * @since       0.3
         * @return      String      HTML Date for settings page
         * @link        http://ednailor.com/2012/04/16/allowing-non-admin-users-to-update-theme-options/
         */
        public function view_settings() 
        {
            
            // quick check on who's viewing ##
            if ( ! current_user_can( Q_SUPPORT_CAPABILITY ) ) { 
                
                wp_die( _e( 'You do not have sufficient permissions to access this page.', $this->text_domain ) );
                
            }
            
            // open wrap ##
            echo '<div class="wrap q_support_wrap">';
            
            // icon and h2 ##
            screen_icon("q_support");
            echo '<h2>'; _e( 'WordPress Support Settings', $this->text_domain ); echo '</h2>';

            // intro blurb ##
            printf( 
                __('
                    <p>You can make support requests to the <a href="%1$s" class="_blank">%2$s</a> or a <a href="%3$s" class="_blank">%4$s</a> or find out more about how the <a href="%5$s" class="_blank">%6$s</a> works.</p>'
                    , $this->text_domain 
                )
                ,esc_url( Q_SUPPORT_DOMAIN.'meta/community-support/' )
                ,esc_html( __("Community", $this->text_domain ) ) 
                ,esc_url( Q_SUPPORT_DOMAIN.'meta/dedicated-agent/' )
                ,esc_html( __("Dedicated Agent", $this->text_domain ) ) 
                ,esc_url( Q_SUPPORT_DOMAIN.'meta/support-process/' )
                ,esc_html( __("Support Process", $this->text_domain ) ) 
            );
            
            // grab the q_support_settings_fields ##
            $q_support_settings_fields = $this->get_q_support_settings_fields();
            
            // grab the saved q_support_settings for the current user ##
            $q_support_settings = $this->get_q_support_settings();
            #wp_die( pr( $q_support_settings ) );
            
?>
            <form id="q_support_settings" name="q_support_settings" method="post">
                <table class="form-table">
                    <tbody>
<?php                        

                        foreach ( $q_support_settings_fields as $field ) {

?>                   
                        <tr valign="top">
                            <th scope="row"><label for="<?php echo $field["name"]; ?>"><?php echo $field["label"]; ?>:</label></th>
                            <td>
                                <select name="q_support_settings_<?php echo $field["name"]; ?>" id="q_support_settings_<?php echo $field["name"]; ?>" class="required">
<?php

                                    foreach ( $field["options"] as $option => $value ) {
                                        
                                        ?><option value="<?php echo $option; ?>" <?php selected( $q_support_settings[$field["name"]], $option ); ?>><?php echo $value; ?></option><?php
                                        
                                    }

?>
                                </select>
                                <p class="description"><?php echo $field["desc"]; ?></p>
                            </td>
                        </tr>
<?php                        

                        } // field loop ##

?> 
                    </tbody>
                </table>
                
                <p><input type="submit" name="submit" id="submit" class="button button-primary" value="<?php _e( "Save Settings", $this->text_domain ); ?>"></p>
                
            </form>


<?php

            
            // tidy up ##
            echo '</div>';
            
        }
        
         
        /**
         * Save user plugin settings
         * 
         * @since       0.5
         * @return      mixed     HTML message on success or False if errors encountered
         */
        public function ajax_q_support_settings() 
        {
            
            // check nonce ##
            check_ajax_referer( 'q_support_nonce', 'nonce' );
            
            // check for "connected" - in case there is nothing to do here ##
            if ( ! $this->get_q_support_user_connected() ) { 
                
                $return = __( "You need to Connect your account first.", $this->text_domain );
                
                // return it ##
                echo json_encode($return);

                // all AJAX calls must die!! ##
                die();
                
            }
            
            if ( ! $this->get_current_user("ID") ) {
                
                $return = __( "User Account Error!", $this->text_domain );
                
                // return it ##
                echo json_encode($return);

                // all AJAX calls must die!! ##
                die();
                
            }
            
            // let's see if the required data was passed ##
            if ( 
                isset( $_POST['q_support_settings_target'] ) && $_POST['q_support_settings_target'] // support audience ##
                #&& isset( $_POST['q_support_settings_agent'] ) && $_POST['q_support_settings_agent'] // dedicated agent ##
                && isset( $_POST['q_support_settings_wordpress'] ) && $_POST['q_support_settings_wordpress'] // send WP settings ##
                && isset( $_POST['q_support_settings_server'] ) && $_POST['q_support_settings_server'] // send server settings ##
                && isset( $_POST['q_support_settings_client'] ) && $_POST['q_support_settings_client'] // send client settings ##
            ) {
                
                // declare a new array ##
                $q_support_settings = array();
                
                // sanitize post data and add to named variables ##
                $q_support_settings['target']= Q_Plugin::sanitize( $_POST['q_support_settings_target'] );
                $q_support_settings['agent'] = Q_Plugin::sanitize( $_POST['q_support_settings_agent'] );
                $q_support_settings['wordpress'] = Q_Plugin::sanitize( $_POST['q_support_settings_wordpress'] );
                $q_support_settings['server'] = Q_Plugin::sanitize( $_POST['q_support_settings_server'] );
                $q_support_settings['client'] = Q_Plugin::sanitize( $_POST['q_support_settings_client'] );
                
                // save it all ##
                $this->set_q_support_settings( $q_support_settings );
                
                // return it ##
                echo json_encode( __( 'Account Settings Saved', $this->text_domain ) );
                
                // all AJAX calls must die!! ##
                die();
                
            } else {
                
                // return it ##
                echo json_encode( __( 'Error Saving Settings!', $this->text_domain ) );
                
                // all AJAX calls must die!! ##
                die();
                
            }

        }
        
               
        /**
         * Register Post Type to store sent support requests and replies ##
         * 
         * @since       0.3
         * @return void
         */
        public function register_post_type() 
        {
            
            // get plugin options ##           
            $q_support_options = $this->get_q_support_options();
            
            $labels = array(
                'name'                      => __('Support Requests', $this->text_domain )
                ,'singular_name'            => __('Support Request', $this->text_domain )
                ,'add_new'                  => __('Ask a Question', $this->text_domain )
                ,'add_new_item'             => __('Add New Requests', $this->text_domain )
                ,'edit_item'                => __('Edit Support Request', $this->text_domain )
                ,'new_item'                 => __('New Support Request', $this->text_domain )
                ,'all_items'                => __('Questions', $this->text_domain )
                ,'view_item'                => __('View Support Request', $this->text_domain )
                ,'search_items'             => __('Search Requests', $this->text_domain )
                ,'not_found'                => __('<p style="margin: 10px;">No Support Requests found <a href="#" class="add-new-h2 q_support_inline">Ask a Question</a></p>', $this->text_domain )
                ,'not_found_in_trash'       => __('No Support Requests found in Trash', $this->text_domain )
                ,'parent_item_colon'        => ''
                ,'menu_name'                => __('Support', $this->text_domain )
            );

            $args = array(
                'labels'                    => $labels
                ,'public'                   => false
                ,'publicly_queryable'       => false
                ,'show_ui'                  => true
                ,'show_admin_column'        => true
                ,'show_in_menu'             => true
                ,'query_var'                => false
                ,'rewrite'                  => array( 'slug' => 'support' )
                ,'capability_type'          => 'post'
                ,'has_archive'              => true
                ,'hierarchical'             => false
                ,'supports'                 => array( 'title', 'editor', 'author', 'thumbnail', 'custom-fields' )
                ,'menu_position'            => 75 // below Tools
                ,'menu_icon'                => $this->get_plugin_url( 'images/icon_wp_support_16_bw.png' ), // 16 x 16 ##
            );

            // add new CPT to WP ##
            register_post_type( 'q_support', $args );
            
            // add flag that CPT added ##
            $q_support_options["cpt"] = true;
            
            // save it all ##
            update_option( 'q_support_options', $q_support_options );
            
        }
        
        
        /**
         * manipulate post row actions ##
         * 
         * @since       0.3
         * @link        http://wordpress.stackexchange.com/questions/14973/row-actions-for-custom-post-types
         */
        public function post_row_actions( $actions, $post )
        {
            
            // check for our post type ##
            if ( $post->post_type == "q_support" ) {
                
                // unset edit ##
                unset( $actions['inline hide-if-no-js'] );
                unset( $actions['edit'] );
                
                // get the question slug ##
                $question_url = get_post_meta( $post->ID, 'q_support_question_url', true );
                
                if ( $question_url ) {
                
                    // add action to view question on wp-support.co ##
                    $actions['view'] = '<a href="'.Q_SUPPORT_DOMAIN.'view/question/'.$question_url.'" class="_blank">View</a>';
                
                }
                
            }
            
            return $actions;
            
        }
        
        
        /**
         * Manage CPT Column ##
         * 
         * @since       0.3
         * @link        https://gist.github.com/bainternet/80a8406c9e714ad6ad37
         * @return      Array   Containing all new column headers
         */
        public function manage_columns() 
        {
            
            // add CPT column class ##
            require_once dirname( __FILE__ ) . '/library/CPT_columns.class.php';
            
            // create an instance
            $CPT_columns = new CPT_columns('q_support');
            
            // remove columns ## 
            $CPT_columns->remove_column('date'); // date ##
            $CPT_columns->remove_column('author'); // author
            $CPT_columns->remove_column('thumbnail'); // thumbnails ##
            
            // add "url" column
            /*
            $CPT_columns->add_column('url',
                array(
                    'label'         => __('URL')
                    ,'type'         => 'text'
                )
            );
            */
            
            // add answers column
            $CPT_columns->add_column('answers',
                array(
                    'label'         => __('Answers')
                    ,'type'         => 'text'
                )
            );
            
            // add comments column
            $CPT_columns->add_column('q_comments',
                array(
                    'label'         => __('Comments')
                    ,'type'         => 'text'
                )
            );
            
            // add views column
            $CPT_columns->add_column('views',
                array(
                    'label'         => __('Views')
                    ,'type'         => 'text'
                )
            );
            
            // add views column
            $CPT_columns->add_column('votes',
                array(
                    'label'         => __('Votes')
                    ,'type'         => 'text'
                )
            );
            
            // add submitted column
            $CPT_columns->add_column('submitted',
                array(
                    'label'         => __('Status')
                    ,'type'         => 'text'
                )
            );
            
            // add asked date column
            $CPT_columns->add_column('asked',
                array(
                    'label'         => __('Asked')
                    ,'type'         => 'text'
                )
            );
            
            // add some filters ##
            #add_filter( 'cpt_columns_text_url', array( $this, 'manage_columns_url' ), 10 ,4 );
            add_filter( 'cpt_columns_text_answers', array( $this, 'manage_columns_answers' ), 10 ,4 );
            add_filter( 'cpt_columns_text_q_comments', array( $this, 'manage_columns_q_comments' ), 10 ,4 );
            add_filter( 'cpt_columns_text_views', array( $this, 'manage_columns_views' ), 10 ,4 );
            add_filter( 'cpt_columns_text_votes', array( $this, 'manage_columns_votes' ), 10 ,4 );
            add_filter( 'cpt_columns_text_submitted', array( $this, 'manage_columns_submitted' ), 10 ,4 );
            add_filter( 'cpt_columns_text_asked', array( $this, 'manage_columns_asked' ), 10 ,4 );
            
        }
        
        
        /**
         * Manage "URL" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_url( $text, $post_id, $column, $column_name )
        {
            
            $question_url = get_post_meta( $post_id, 'q_support_question_url', true );
            
            return 
                ( $question_url ) ? // do we have an URL ## 
                sprintf( 
                    '<a href="%s" class="_blank">%s</a>'
                    , esc_url( Q_SUPPORT_DOMAIN.'view/question/'.$question_url )
                    , esc_html( __("Open", $this->text_domain ) ) 
                ) : // if so return the link ##
                '-' ; // or a dead space ##
            
        }
        
        
        /**
         * Manage "Answers" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_answers( $text, $post_id, $column, $column_name )
        {
            
            return get_post_meta( $post_id, 'q_support_question_answers' ) ? get_post_meta( $post_id, 'q_support_question_answers', true ) : 0;
            
        }
        
        
        /**
         * Manage "Comments" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_q_comments( $text, $post_id, $column, $column_name )
        {
            
            return get_post_meta( $post_id, 'q_support_question_comments' ) ? get_post_meta( $post_id, 'q_support_question_comments', true ) : 0;
            
        }
        
        
        /**
         * Manage "Views" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_views( $text, $post_id, $column, $column_name )
        {
            
            return get_post_meta( $post_id, 'q_support_question_views' ) ? get_post_meta( $post_id, 'q_support_question_views', true ) : 0;
            
        }
        
        
        /**
         * Manage "Votes" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_votes( $text, $post_id, $column, $column_name )
        {
            
            // get number of votes ##
            $votes_up = get_post_meta( $post_id, 'q_support_question_votes_up' ) ? (int)get_post_meta( $post_id, 'q_support_question_votes_up', true ) : 0 ;
            $votes_down = get_post_meta( $post_id, 'q_support_question_votes_down' ) ? (int)get_post_meta( $post_id, 'q_support_question_votes_down', true ) : 0 ;
            
            // kick back formatted votes ##
            return sprintf(
                "%s %s"
                , '<span class="wps_icons thumb_up" title="'.__("Up Votes", $this->text_domain ).'">'.$votes_up.'</span>'
                , '<span class="wps_icons thumb_down" title="'.__("Down Votes", $this->text_domain ).'">'.$votes_down.'</span>'
            );
            
        }
        
        
        /**
         * Manage "Synced" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_submitted( $text, $post_id, $column, $column_name )
        {
            
            $q_support_submitted = get_post_meta( $post_id, "q_support_submitted", true );
            
            return ( $q_support_submitted ) ? '<span class="wps_icons flag_green" title="'.__("Connected to WP Support").'"></span>' : '<span class="flag_red" title="'.__("Not connected to WP Support").'"></span>' ;
            
        }
        
        
        /**
         * Manage "Asked" Column ##
         * 
         * @since       0.3
         * @return      String   HTML
         */
        public function manage_columns_asked( $text, $post_id, $column, $column_name )
        {
            
            // human difference ##
            #$time_since_posted = human_time_diff( get_the_time( 'U', $post_id ), current_time( 'timestamp' ) ).' '.__("ago", $this->text_domain);

            #return get_the_time( 'U', $post_id );
            return $this->human_time_diff( get_post_time('G', true, $post_id ) );
            
        }
        
        
        /**
         * Manage Users columns
         * 
         * @since       0.7
         * @return      Array 
         */
        public function manage_users_columns( $column ) 
        {
            
            $column['wps_key'] = 'WPS Key';
            
            return $column;
            
        }

        
        /**
         * Print Users Column Data
         *      
         * @since       0.7
         * @return      void 
         */
        public static function manage_users_custom_column( $val, $column_name, $user_id ) 
        {
            
            // grab each users meta ##
            $q_support_settings = get_user_meta( $user_id, 'q_support_settings', true );
            
            switch ( $column_name ) {
                
                case 'wps_key' :
                    
                    return $q_support_settings && isset ( $q_support_settings["key"] ) ? $q_support_settings["key"] : '-';
                    break;

            }

        }
        
        
        /**
         * Limit the results displayed to the current user
         * 
         * @since       0.4
         * @return      void
         */
        public function limit_results()
        {
            
            // load the scripts on only the plugin admin page 
            if ( $this->get_current_post_type() == 'q_support' ) {
            
                // add class ##
                require_once dirname( __FILE__ ) . '/library/Q_Limit_To_Current_User.class.php';

                // quick check ##
                if ( ! class_exists ( 'Q_Limit_To_Current_User' ) ) { return false; }

                // add the action to filter ##
                $q_limit_to_current_user = New Q_Limit_To_Current_User();
                add_action( 'pre_get_posts', array ( $q_limit_to_current_user, 'query_set_only_author' ) );

            }
            
        }
        
        
        /**
         * Get Current Post Type
         * 
         * @since       0.8 
         * @global      type $post
         * @global      type $typenow
         * @global      type $current_screen
         * @return      Mixed       String Post Type Name OR boolean false
         */
        public function get_current_post_type() 
        {

            global $post, $typenow, $current_screen;

            // we have a post so we can just get the post type from that
            if ( $post && $post->post_type ){
                return $post->post_type;
            }

            // check the global $typenow - set in admin.php
            elseif( $typenow ) {
                return $typenow;
            }

            // check the global $current_screen object - set in sceen.php
            elseif( $current_screen && $current_screen->post_type ) {
                return $current_screen->post_type;
            }

            //lastly check the post_type querystring
            elseif( isset( $_REQUEST['post_type'] ) ) {
                return sanitize_key( $_REQUEST['post_type'] );
            }

            // we do not know the post type!
            return null;

        }
        
        
        /**
         * Add notice to WP admin if plugin is upgraded ##
         * 
         * @since       0.3
         * @return      void
         */
        public function admin_update_notice() 
        {
            
            // get plugin options ##           
            $q_support_options = $this->get_q_support_options();
            
            // check the plugin version ##
            $version = isset( $q_support_options["version"] ) ? $q_support_options["version"] : Q_SUPPORT_VERSION;
            if ( $version != Q_SUPPORT_VERSION ) {
                
                 // update stored version number ##
                $q_support_options["version"] = Q_SUPPORT_VERSION;
                
                // store notice for later use ##
                $q_support_options["deferred_admin_notices"][] = __( "WordPress Support: Upgraded version $version to ".Q_SUPPORT_VERSION.".", $this->text_domain );
                
            }
            
            // save it all ##
            update_option( 'q_support_options', $q_support_options );
        
        }
        
        
        /**
         * Add notices to WP admin
         * 
         * @since       0.3
         */
        public function admin_notices() 
        {
            
            // get plugin name ##
            #if ( isset( $_GET['page'] ) && ( $_GET['page'] == 'q_support' ) ) {
            
                // get plugin options ##           
                $q_support_options = $this->get_q_support_options();
                
                // grab notices or return an empty array ##
                $notices = isset( $q_support_options["deferred_admin_notices"] ) ? $q_support_options["deferred_admin_notices"] : array();
                
                // Suppress "Plugin activated" notice.
                if ( isset( $q_support_options["deferred_admin_notices"] ) ) { unset( $_GET['activate'] ); }
                
                // check for notices in the the querstring with $_GET ##
                if ( isset( $_GET["settings-updated"] ) ) {

                    if ( isset ( $_GET["page"] ) && $_GET["page"] === 'settings' ) {

                        $notices[] = __("Support Settings Saved.", $this->text_domain); // add saved message ##

                    }

                }

                // do we have some notices ?
                if ( $notices ) {

                    // loop over each notice ##
                    foreach ( $notices as $notice ) {
                        #if ( !$notice ) continue; // skip over empty values ##
                        echo "<div class='updated'><p>$notice</p></div>";
                    }

                }

                // tidy up notices ##

                // delete them, so we don't do this again ##
                unset( $q_support_options["deferred_admin_notices"] );

                // save it all ##
                update_option( 'q_support_options', $q_support_options );
                
            #}
            
        }
        
        
        /**
         * Load Text Domain for translations ##
         * 
         * @since       0.3
         * 
         * @link        http://geertdedeckere.be/article/loading-wordpress-language-files-the-right-way ##
         */
        public function load_plugin_textdomain() 
        {
            
            load_plugin_textdomain( $this->text_domain, FALSE, dirname(plugin_basename(__FILE__)).'/languages/' );
            
        }
        
        
        /**
         * Get Plugin URL
         * 
         * @since       0.3
         * @return      string  Absoulte URL to plugin directory
         */
        public static function get_plugin_url( $path = '' ) 
        {
            
            return plugins_url( ltrim( $path, '/' ), __FILE__ );
            
        }
        
        
        /*
         * Add settings link on plugin page ##
         * 
         * @since       0.3
         * @src         http://bavotasan.com/2009/a-settings-link-for-your-wordpress-plugins/
         * @return      Array   Links to show on plugin option
         */
        public function add_plugin_settings_link($links) 
        { 
            
            $settings_link = '<a href="'.$this->plugin_url.'">'.__( "Settings", $this->text_domain ).'</a>'; 
            array_push( $links, $settings_link ); 
            return $links; 
            
        }
 
        
        /**
         * Enqueue Plugin Scripts & Styles
         * 
         * @since       0.2
         * @return      void
         */
        public function admin_enqueue_scripts() 
        {
            
            wp_register_style( 'q-support-css', $this->get_plugin_url( 'css/q-support.css' ) );
            wp_enqueue_style( 'q-support-css' );
            wp_enqueue_script( 'q-support-js', $this->get_plugin_url( 'javascript/q-support.js' ), array( 'jquery' ), Q_SUPPORT_VERSION, true );
            wp_enqueue_script( 'jquery-validate-js', $this->get_plugin_url( 'javascript/jquery.validate.min.js' ), array( 'jquery' ), '1.11.1', false );

            $params = array(
                'ajax_nonce' => wp_create_nonce( 'q_support_nonce' ),
            );
            wp_localize_script( 'q-support-js', 'ajax_object', $params );
            
        }
        
        
        /**
         * Enqueue Plugin Scripts & Styles
         * 
         * @since       0.5
         * @return      void
         */
        public function wp_enqueue_scripts() 
        {
            
            wp_register_style( 'q-support-css', $this->get_plugin_url( 'css/q-support.css' ) );
            wp_enqueue_style( 'q-support-css' );
            
        }
        
        
        /**
         * Debug all Getter functions
         * 
         * @since       0.3
         * @return      string  HTML content of debug info ##
         */
        public function view_q_support_data()
        {
            
            // quick check on who's viewing ##
            if ( ! current_user_can( Q_SUPPORT_CAPABILITY ) ) { 
                
                wp_die( _e( "You do not have sufficient permissions to access this page.", $this->text_domain ) );
                
            }
            
            // add kint for pretty var_dump ##
            require_once dirname( __FILE__ ) . '/library/kint/Kint.class.php';
            
            // open wrap ##
            echo '<div class="wrap q_support_wrap">';
            
            // icon and h2 ##
            screen_icon("q_support");
            echo '<h2>'; _e("Plugin Data"); echo '</h2>';
            
            // notice space ##
            echo '<div class="q_support_response updated"><p></p></div>';
            
            // intro blurb ##
            printf( 
                '<p>This screen allows you to view all the data stored in the database which can be submitted allong with support requests - <a href="%1$s" class="_blank">%2$s</a></p>'
                ,esc_url( Q_SUPPORT_DOMAIN.'plugin/question-data/' )
                ,esc_html( __("Documentation", $this->text_domain ) ) 
            );
            
            d( $this->get_q_support_options() );
            d( $this->get_q_support_settings() ); // this is now a personal setting - per user, so it is still relevant here ?
            d( $this->get_server() );
            d( $this->get_client() );
            d( $this->get_htaccess() );
            d( $this->get_wp_config() );
            d( $this->get_wordpress() );
            d( $this->get_post_types() );
            d( $this->get_theme() );
            d( $this->get_plugins() );
            d( $this->get_sidebars() );
            
            // link to error log ##
            printf( 
                '<p><a href="%s" class="q_support_data_clear button">%s</a></p>'
                ,esc_url( '#' )
                ,esc_html( __("Reload Stored Data", $this->text_domain ) )
            );
            
            // link to error log ##
            printf( 
                '<p>%s</p>'
                ,esc_html( __("Note: User and plugin activation data is only cleared when the plugin is uninstalled.", $this->text_domain ) )
            );
            
            // close wrap ##
            echo '</div>';
            
        }
        
        
        /*
        * AJAX Clear Support Data ##
        * 
        * @since        0.3
        *
        * @return       string  HTML notification message ##  
        **/
        public function ajax_q_support_data_clear() 
        {
            
            // quick check on who's viewing ##
            if ( ! current_user_can( Q_SUPPORT_CAPABILITY_ADMIN ) ) { 
                wp_die( _e( "You do not have sufficient permissions to access this page.", $this->text_domain ) );
            }
            
            // check nonce ##
            check_ajax_referer( 'q_support_nonce', 'nonce' );
            
            // array of plugin options ##
            $options = array (
                #'q_support_options' // plugin options ##
                'q_support_settings' // settings ##
            );
            
            // remove plugin options ##
            $this->delete_options( $options );
            
            // clear all transient data ##
            $this->delete_cache();
            
            // delete the error log file ##
            $this->delete_error_log();
            
            // delete CPT posts ##
            global $wpdb;

            $query = "
                DELETE FROM wp_posts 
                WHERE post_type = 'q_support' 
            ";

            #LIMIT 0, 3000
            
            $wpdb->query($query);
            
            // now reload the data ##
            #$this->get_q_support_options();
            #$this->get_q_support_user();
            
            // reload all other data ##
            $this->get_q_support_settings(); // this is now a personal setting - per user, so it is still relevant here ?
            $this->get_server();
            $this->get_client();
            $this->get_htaccess();
            $this->get_wp_config();
            $this->get_wordpress();
            $this->get_post_types();
            $this->get_theme();
            $this->get_plugins();
            $this->get_sidebars();
            
            $return = sprintf( 
                __(
                    'Plugin Data Reloaded - <a href="%s">%s</a> to see the new data below.'
                    ,$this->text_domain
                )
                , esc_url( $this->plugin_url_data )
                , esc_html( __("Refresh", $this->text_domain ) ) 
            );

            // return it ##
            echo json_encode($return);

            // all AJAX calls must die!! ##
            die();
            
	}
        
        
        /**
         * Add a Screen Options panel to the plugin settings page
         * 
         * @since       0.3
         * @link        http://pippinsplugins.com/add-screen-options-tab-to-your-wordpress-plugin/
         * @link        http://w-shadow.com/blog/2010/06/29/adding-stuff-to-wordpress-screen-options/
         * @return      void
         */
        public function add_screen_options() 
        {

            global $q_support_page;

            $screen = get_current_screen();

            // get out of here if we are not on our settings page
            if( ! is_object($screen) || $screen->id != $q_support_page ) {
                return;
            }
            
            $args = array(
                'label' => __('Question Per Page', $this->text_domain),
                'default' => 10,
                'option' => 'x_per_page'
            );
            
            add_screen_option( 'per_page', $args );
                
        }
        
        
        /**
         * Add Help Tab
         * 
         * @since        0.3
         * @link         https://codex.wordpress.org/Function_Reference/add_help_tab ##
         * @return       void
         */
        public function add_help_tab() 
        {
            
            $this->help_tabs = array(
            
                // The assoc key represents the ID cannot contain spaces ##
                'support' => array(
                    'title'   => __( 'Support', $this->text_domain )
                )

            );
            
            $help_tabs = get_current_screen()->get_help_tabs();
            get_current_screen()->remove_help_tabs();

            foreach( $this->help_tabs as $id => $data ) {
                get_current_screen()->add_help_tab( array (
                        'id' => $id
                    ,   'title' => __( $data['title'], 'q_support' )
                    ,   'callback' => array( $this, 'q_support_form' )
                ));
            }
                
            if ( count( $help_tabs ) ) {
                foreach ( $help_tabs as $help_tab ) {
                    get_current_screen()->add_help_tab( $help_tab );
                }
            }
            
	}
        
        
        /*
        * Add Support Form to Help Tab ##
        * 
        * @since        0.3
        *
        * @return       string  HTML code for help tab ##  
        **/
        public function q_support_form( $screen, $tab ) 
        {
            
            // is the user added correctly already ? ##
            if ( ! $this->get_q_support_user_connected() ) {
  
                // notice space ##
                echo '<div class="q_support_response updated"><p></p></div>';
                
?>
                <p>
                <?php 
                printf( 
                    __(
                        'To submit support requests, you first need to <a href="#" class="button q_support_connect q_support_space_both">%s</a> your account - <a href="%s">%s</a>'
                        ,$this->text_domain
                    )
                    #,esc_url( $this->plugin_url_connect ) // connect to API ##
                    ,esc_html( __("Connect", $this->text_domain ) ) 
                    ,esc_url( Q_SUPPORT_DOMAIN.'meta/terms-of-use/' ) // terms ##
                    ,esc_html( __("Terms of Use", $this->text_domain ) ) 
                ); ?>
                </p>
<?php
                
            // all good - let's show the form ##
            } else {
            
?>
            <p><?php _e("Complete the following form to request support for a problem you are having.", $this->text_domain); ?></p>
            <form id="q_support_request" name="q_support_request" method="post">
                
                <div class="q_support_response updated"><p></p></div>
                
                <table class="form-table">
                    <tbody>
                        
                        <tr valign="top">
                            <th scope="row"><label for="title"><?php _e("Support title", $this->text_domain) ?>:</label></th>
                            <td><input name="q_title" type="text" id="q_title" value="" class="large-text required"></td>
                        </tr>
                        
                        <tr valign="top">
                            <th scope="row"><label for="title"><?php _e("Describe the problem", $this->text_domain) ?>:</label></th>
                            <td><textarea name="q_question" rows="6" cols="30" id="q_question" class="large-text code required" style="resize: vertical;"></textarea></td>
                        </tr>
                        
                        <tr valign="top">
                            <th scope="row"><label for="url"><?php _e("Problem URL", $this->text_domain) ?>:</label></th>
                            <td><input name="q_url" type="text" id="q_url" value="<?php echo $this->get_current_url(); ?>" class="large-text required"></td>
                        </tr>
                        
                        <tr valign="top">
                            <th scope="row"><label for="url"><?php _e("Category", $this->text_domain) ?>:</label></th>
                            <td>
                                <select name="q_category" id="q_category">
<?php

                                    foreach ( $this->question_categories as $category ) {
                                        
                                        echo "<option value='{$category}'>{$category}</option>";
                                        
                                    }

?>
                                </select>
                            </td>
                        </tr>
                        
                    </tbody>
                </table>
                
                <p class="q_support_submit"><input type="submit" name="submit" id="submit" class="button button-primary" value="Get Support"></p>
                
            </form>
            <hr/>
            <p>
                <?php printf( 
                    '<a href="%s" class="button q_support_space_right">%s</a>'
                    , esc_url( $this->plugin_url_settings )// link to settings ##
                    , esc_html( __("View Settings", $this->text_domain ) ) 
                );  ?>
                <?php _e("Be careful to never share your password or confidential information with anyone that you do not trust!", $this->text_domain); ?>
            </p>
<?php
            
            } // check for user connect ##

	}
        
        
        /*
        * AJAX Capture Support Form Request
        * 
        * @since        0.3
        * @return       string      HTML for help tab feedback
        **/
        public function ajax_q_support_form() 
        {
            
            // check nonce ##
            check_ajax_referer( 'q_support_nonce', 'nonce' );
            
            // get user ##
            $user = $this->get_current_user();
            
            // validate user ##
            if ( ! $user ) { 
            
                // return it ##
                echo json_encode( __( 'User Account Error', $this->text_domain ) );

                // all AJAX calls must die!! ##
                die();
            
            } 
            
            // get user plugin settings ##
            $q_support_settings = $this->get_q_support_settings();
            
            // let's see if the required data was passed ##
            if ( 
                isset( $_POST['q_title'] ) && $_POST['q_title'] // question title ##
                && isset( $_POST['q_question'] ) && $_POST['q_question'] // question itself ##
                && isset( $_POST['q_url'] ) && $_POST['q_url'] // question URL ##
                && isset( $_POST['q_category'] ) && $_POST['q_category'] // question Category ##
            ) {
                
                // sanitize post data and add to named variables ##
                $title = Q_Plugin::sanitize( $_POST['q_title'], 'stripslashes', true );
                $question = Q_Plugin::sanitize( $_POST['q_question'], 'stripslashes', true );
                $category = Q_Plugin::sanitize( $_POST['q_category'], 'stripslashes', true );
                
                // question URL ##
                $url = Q_Plugin::sanitize( $_POST['q_url'], 'stripslashes', true );
                
                // returns absolute path to dump file on this WP install ##
                $dump = $this->get_plugin_url().'/dump/'.$this->create_dump( $this->generate_key() );
                
                // question security ##
                $agent = false; 
                $secure = false;
                if ( $q_support_settings["target"] == 'agent' && $q_support_settings["agent"] ) {
                
                    $agent = (int)Q_Plugin::sanitize( $q_support_settings["agent"], 'integer' );
                    $secure = 1; // yes - we need to secure this ##
                
                }
                
                // Save Support Request to WP ##
                $request = array(
                        'post_title'       => $title
                    ,   'post_type'        => 'q_support'
                    ,   'post_content'     => $question
                    ,   'post_status'      => 'publish'
                    ,   'post_author'      => $user->ID
                );

                // Insert the post into the database
                $post_id = wp_insert_post( $request );
                
                // save all that lovely post meta also ##
                add_post_meta( $post_id, 'q_support_url', $url ); // URL ##
                add_post_meta( $post_id, 'q_support_dump', $dump ); // dump ##
                if ( $secure ) add_post_meta( $post_id, 'q_support_secure', $secure ); // secure ##
                if ( $agent ) add_post_meta( $post_id, 'q_support_agent', $agent ); // agent ##
                add_post_meta( $post_id, 'q_support_category', $category ); // category ##
                add_post_meta( $post_id, 'q_support_settings', $q_support_settings ); // save plugin settings at time of submission ##
                add_post_meta( $post_id, 'q_support_submitted', 'false' ); // sync flag - default to false ##
                
                // now to submit the question to the API -------------------------- ##
                
                // check details ##
                if ( $user && is_user_logged_in() && $user->user_login && $user->user_email ) {
                    
                    // grab key from the user's meta data ##
                    $q_support_key = $this->get_q_support_user_key();
                    
                    // no key - throw an error ##
                    if ( ! $q_support_key ) {
                        
                        // kick back out ##
                        echo json_encode( __( 'User Account Key Error', $this->text_domain ) );

                        // all AJAX calls must die!! ##
                        die();
                        
                    }
                    
                    // store it ##
                    Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__key' => $q_support_key );
                    
                    // let's dump and check some things here ##
                    $q_dump = array(
                            'title'     => $title
                        ,   'question'  => $question
                        ,   'category'  => isset ( $category ) ? $category : __( 'General', $this->text_domain )
                        ,   'tag'       => isset( $tag ) ? $tag : __( 'General', $this->text_domain )
                        ,   'url'       => $url
                        ,   'key'       => $q_support_key
                        ,   'user'      => $user->ID
                        ,   'dump'      => $dump
                        ,   'secure'    => $secure
                        ,   'agent'     => $agent
                    );

                    // store it ##
                    if ( Q_DEBUG ) { update_option( "q_support_question_wps", $q_dump ); }
                    
                    // wp_remote_post call to API ## 
                    $post_data = array( 
                        'method'        => 'set_question'
                        ,'key'          => $q_support_key
                        ,'title'        => $title
                        ,'question'     => $question
                        ,'category'     => isset ( $category ) ? $category : __( 'General', $this->text_domain )
                        ,'tag'          => isset( $tag ) ? $tag : __( 'General', $this->text_domain )
                        ,'url'          => $url // question URL ##
                        ,'dump'         => $dump // dump file ##
                        ,'secure'       => $secure // is it? ##
                        ,'agent'        => $agent // who to? ##
                    );

                    // http://kovshenin.com/2011/a-note-on-wp_http-and-http-requests-in-wordpress/
                    $headers    = array( 
                        'Authorization' => 'Basic ' . base64_encode( "$user->user_login:$user->user_pass" ) // very basic ##
                    );

                    // do it ! ##
                    $result     = wp_remote_post ( 
                                    $this->q_support_api_url 
                                    ,array( 
                                            'headers'   => $headers
                                        ,   'body'      => $post_data
                                        ,   'timeout'   => $this->q_support_api_timeout, // set timeout ##
                                    ) 
                                );
                    
                    #echo print_r ( $result["body"] );
                    #die();
                    
                    // store it ##
                    Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__result' => var_export( $result, true ) );

                    if ( ! $result ) {

                        // store it ##
                        Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__result' => 'empty' );
                        
                        // prepare notice message ##
                        $return = __("Unknown Error");
                    
                    } else if ( is_wp_error( $result ) ) {

                        // grab error ##
                        $error_message = $result->get_error_message();
                        
                        // save to the post ##
                        add_post_meta( $post_id, 'q_support_response', $error_message ); // response ##
                        
                        // store it ##
                        Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__response' => $error_message );
                        
                        // prepare notice message ##
                        $return = $error_message;

                    } else {

                        // decode the results ##
                        $results = isset( $result["body"] ) ? json_decode( $result["body"] ) : wp_die( $result ) ;

                        // save API results ##
                        add_post_meta( $post_id, 'q_support_response', isset ( $results->response ) ? $results->response : '' ); // response ##
                        add_post_meta( $post_id, 'q_support_debug', isset( $results->debug ) ? $results->debug : '' ); // debug ##
                        add_post_meta( $post_id, 'q_support_code', isset( $result["response"] ) ? $result["response"] : '' ); // api_code ##

                        // if user was created - update the "connected" and move on ##
                        if ( isset( $results->status ) && $results->status === true ) {
                            
                            // last check - for returned key ##
                            if ( ! isset( $results->key ) ) {

                                // grab error ##
                                $error_message = $result->get_error_message();

                                // save to the post ##
                                add_post_meta( $post_id, 'q_support_response', isset ( $results->response ) ? $results->response : '' ); // response ##
                                
                                // store it ##
                                Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__response' => isset( $results->response ) ? $results->response : '' );
                                
                                // prepare notice message ##
                                $return = $error_message;

                            } else {
                                
                                // store it ##
                                Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__question_key' => isset ( $results->key ) ? $results->key : '' );
                                
                                // save the question key in post_meta ##
                                add_post_meta( $post_id, 'q_support_question_key', isset ( $results->key ) ? $results->key : '', true );

                                // update post meta flag to TRUE ##
                                update_post_meta( $post_id, 'q_support_submitted', isset ( $results->status ) ? 'true' : 'false', 'false' );

                                // get question URL ##
                                $question_url = isset( $results->url ) ? $results->url : '' ;
                                add_post_meta( $post_id, 'q_support_question_url', $question_url, true );

                                // notice // add the question link ##
                                $return = $results->response.' - '.sprintf( 
                                            __( '<a href="%s"target="_blank">%s</a>', $this->text_domain )
                                            ,esc_url( Q_SUPPORT_DOMAIN.'view/question/'.$question_url )
                                            ,esc_html( __("View the Question", $this->text_domain ) ) 
                                        );

                            }
                                
                        } else { // status came back false ##
                            
                            // store it ##
                            Q_Plugin::$store["ajax_q_support_form"] = array( 'q_support_form__status' => 'false' );
                            
                            // just send back the API response ##
                            $return = isset( $results->response ) ? $results->response : __( 'Error!' );

                        }

                    }


                } else {
                    
                    // store it ##
                    Q_Plugin::$store[] = array( 'key' => 'q_store', 'value' => array( 'q_support_form__missing_user' => 'true' ) );
                    
                    $return = __( 'Error - Missing User Data.', $this->text_domain );
                    
                }
                
            } else {
                
                // store it ##
                Q_Plugin::$store[] = array( 'key' => 'q_store', 'value' => array( 'q_support_form__no_data_passed' => 'true' ) );
                
                $return = __( 'Error - No data passed.', $this->text_domain );

            }

            // return it ##
            echo json_encode( $return );

            // all AJAX calls must die!! ##
            die();
            
	}
        
        
        /**
         * Set Error Handlers
         * 
         * @since       0.3
         * @link        http://us1.php.net/set_error_handler
         * @ return     void
         */
        public function set_error_logging()
        {
            
            // If WP_DEBUG is set to true, let's guess the user knows what they are doing and turn this feature off ##
            if ( WP_DEBUG === true ) {
                return false;
            }
            
            // create the error log file ##
            $this->make_error_log();
            
            // We need to make sure we can capture errors ##
            error_reporting( E_ALL );
            ini_set( 'error_reporting', E_ALL );
            ini_set( 'display_errors', FALSE );
            ini_set( 'display_startup_errors', FALSE );
            ini_set( 'html_errors', FALSE );
            
            // Error handler ##
            set_error_handler( array( $this, 'log_error' ) );
            
            // Uncaught exception handler ##
            set_exception_handler( array( $this, 'log_exception' ) );
            
            // Fatal ones !! ##
            register_shutdown_function( array( $this, 'log_shutdown' ) );
            
        }
        
        
        /**
         * Check Error Log ##
         * 
         * @since       0.3
         * @return      Array     data about error log ##
         */
        private function check_error_log() 
        {
            
            $check_error_log = array();
            $check_error_log["filesize"] = 0; // default to zero ##
            $check_error_log["file_exists"] = false; // default to no ##
            
            if ( file_exists( Q_SUPPORT_LOG_FILE ) ) {
                
                // ok ! ##
                $check_error_log["file_exists"] = true;

                // grab size ##
                $check_error_log["filesize"] = round( filesize( Q_SUPPORT_LOG_FILE ) / 1024 );
                
            }
            
            if ( is_readable( Q_SUPPORT_LOG_FILE ) ) {
                $check_error_log["is_readable"] = true;
            }
            
            if ( is_writable( Q_SUPPORT_LOG_FILE ) ) {
                $check_error_log["is_writable"] = true;
            }
            
            
            
            // archive file, if it's too big ##
            if ( $check_error_log["filesize"] > Q_SUPPORT_LOG_MAX_KB ) {
                
                $this->rename_error_log();
                
            }
            
            return $check_error_log;
            
        }
        
        
        /**
         * Make Error Log ##
         * 
         * @since       0.3
         * @return      void
         */
        private function make_error_log() 
        {
            
            // get error log details ##
            $check_error_log = $this->check_error_log();
                    
            if ( $check_error_log["file_exists"] === false ) {
                
                $make_error_log = @fopen( Q_SUPPORT_LOG_FILE, 'w' );
                
                if ( $make_error_log === false ) {
                    
                    // get plugin options ##           
                    $q_support_options = $this->get_q_support_options();
            
                    // save error ##
                    $q_support_options["make_error_log"] = false; 

                    // save it all ##
                    update_option( 'q_support_options', $q_support_options );
                    
                }
                
            }

        }
        
        
        /**
         * Rename Error Log ##
         * 
         * @since       0.3
         * @return      boolean     True or False ##
         */
        private function rename_error_log() 
        {
            
            // get todays date ##
            $now = date( "Y-m-d-H-i-s", strtotime( "now" ));
            
            // get plugin options ##           
            $q_support_options = $this->get_q_support_options();

            // save error ##
            $q_support_options["rename_error_log"] = $now; 

            // save it all ##
            update_option( 'q_support_options', $q_support_options );
            
            // do some renaming ##
            return @rename( Q_SUPPORT_LOG_FILE, Q_SUPPORT_LOG_PATH.$now.'_debug.log' );
            
        }        
        
        
        /**
         * Delete Error Log ##
         * 
         * @since       0.3
         * @return      void
         */
        private static function delete_error_log() 
        {
            
            #@unlink( Q_SUPPORT_LOG_FILE );
            
            $files = glob( Q_SUPPORT_LOG_PATH.'*' ); // get all log files ##
            foreach ( $files as $file ){ // iterate files ##
                if ( is_file( $file ) ) {
                    @unlink( $file ); // delete file ##
                }
            }
            
        }        
        
        
        /*
         * Error handler, passes flow over to the exception logger with new ErrorException.
         * 
         * @since       0.3
         * @return      void
         */
        public function log_error( $errno, $errstr, $errfile, $errline, $errcontext = null ) 
        {
            
            $l = error_reporting();
            
            if ( $l & $errno ) {
                
                $exit = false;
                switch ( $errno ) {
                    
                    case E_USER_ERROR:
                    case E_ERROR:
                        $type = 'Fatal Error';
                        $exit = true;
                    break;
                    case E_USER_WARNING:
                    case E_WARNING:
                        $type = 'Warning';
                    break;
                    case E_USER_NOTICE:
                    case E_NOTICE:
                    case @E_STRICT:
                        $type = 'Notice';
                    break;
                    case @E_RECOVERABLE_ERROR:
                        $type = 'Catchable';
                    break;
                    default:
                        $type = 'Unknown Error';
                        $exit = true;
                    break;
                
                }
                
                $this->log_exception ( new \ErrorException( $type.': '.$errstr, 0, $errno, $errfile, $errline ) );
                
            }
            
        }
        
        
        /**
         * Save sanitized uncaught exception handlers to log ##
         * 
         * @since       0.3
         * 
         * @return      void
         */
        public function log_exception( Exception $e ) 
        {
            
            $message = "[".date(Q_SUPPORT_DATE_FORMAT)."] PHP {$e->getMessage()} in {$e->getFile()} @ line {$e->getLine()};";
            file_put_contents( Q_SUPPORT_LOG_FILE, $message . PHP_EOL, FILE_APPEND );

            // Don't execute PHP internal error handler ##
            return false;
           
        }
        
        
        /**
         * Save fatal errors to log / database ##
         * 
         * @since       0.3
         * @src         http://stackoverflow.com/questions/4410632/handle-fatal-errors-in-php-using-register-shutdown-function
         * 
         * @return      void
         */
        public function log_shutdown() 
        {
            
            $error = error_get_last();
            
            if ( $error["type"] == E_ERROR ||  $error["type"] == E_USER_ERROR ) {
            
                $this->log_error( $error["type"], $error["message"], $error["file"], $error["line"] );
                
                // Don't execute PHP internal error handler ##
                exit;
                
            }
            
        }
        
        
        /**
         * Get Error Log contents ##
         * 
         * @since       0.3
         * @return      array   Containing sanitized content from error log ##
         */
        public function view_error_log( $lines = 21, $length = 60, $file = null ) 
        {
            
            // quick check on who's viewing ##
            if ( ! current_user_can( Q_SUPPORT_CAPABILITY ) ) { 
                
                wp_die( _e( "You do not have sufficient permissions to access this page.", $this->text_domain ) );
                
            }
            
            if ( !$file ) { $file = Q_SUPPORT_LOG_FILE; } // default to plugin error log file ##
            if ( !$lines ) { $lines = 21; } // lines backup - WHY?? ##
            
            // we double the offset factor on each iteration ##
            $multiplier = 1;
            
            // add slashes -- required for winNT? ##
            #$file = addslashes( $file );
            
            // get the size of the file ##
            $bytes = file_exists( $file ) ? filesize( $file ) : 0; 
            
            #wp_die( $bytes. ' / ' . $file );
            
            // not yet complete ##
            $complete = FALSE;
            
            // start a new array ##
            $log = array();
            
            // open it up ##
            $fp = fopen( $file, "r" ) or wp_die( _e( "Can't open $file", $this->text_domain ) );
            
            while ( $complete === FALSE ) {
                
                //seek to a position close to end of file
                $offset = ( (int)$lines * (int)$length * (int)$multiplier );
                #echo "offset: {$offset} / lines: {$lines} / length {$length} / multiplier {$multiplier} <br />";
                fseek( $fp, -$offset, SEEK_END );

                // we might seek mid-line, so read partial line
                // if our offset means we're reading the whole file, 
                // we don't skip...
                if ( $offset < $bytes ) {
                    fgets( $fp );
                }
                
                // read all following lines, store last x ##
                while ( !feof( $fp ) ) {
                    
                    $line = fgets( $fp );
                    #echo($line);
                    array_push( $log, $line );
                    #$log[] = $line;
                    if ( count( $log ) > $lines ) {
                        array_shift( $log );
                        $complete = TRUE;
                    }
                    
                }
                
                #echo count($log);
                
                // if we read the whole file, we're done, even if we don't have enough lines ##
                if ( $offset >= $bytes ) {
                    $complete = TRUE;
                } else {
                    $multiplier *= 2; //otherwise let's seek even further back
                }
                
            }
            
            // tidy up ##
            fclose( $fp );
            
            // flip the error log - perhaps a bit controversial !! ##
            $log = array_reverse($log);
            
            // format to print ##
            $log_format = '';
            
            // counter ##
            $count = 0;
            
            // clean up the array ##
            foreach( $log as $key => $value ) {
                if( is_null($value) || $value == '' || $value === false ) {
                    unset( $log[$key] );
                } else {
                                       
                    $log_format .= "<p class='q_support_log_format'><span class='q_support_log_count'>$count</span>{$value}</p>";
                    
                }
                $count++;
            }
            
            // open wrap ##
            echo '<div class="wrap q_support_wrap">';
            
            // icon and h2 ##
            screen_icon("q_support");
            echo '<h2>'; _e("PHP Error Log"); echo '</h2>';
            
            // intro blurb ##
            printf( 
                '<p>Here are the last 20 entries ( in reverse order ) from the PHP Error Log, you can view the entire file using the link at the bottom - <a href="%1$s" class="_blank">%2$s</a></p>'
                ,esc_url( Q_SUPPORT_DOMAIN.'plugin/error-log/' )
                ,esc_html( __("Documentation", $this->text_domain ) ) 
            );
            
            // dump it ##
            echo ( $log_format );
            
            // link to error log ##
            printf( 
                '<a href="%s" class="q_support_log_button button _blank">%s ( %s )</a>.'
                ,esc_url( $this->get_plugin_url( '/logs/debug.log' ) )
                ,esc_html( __("View Full Error Log", $this->text_domain ) )
                ,esc_html( round( $bytes / 1024 )." kb" )
            );
            
            // close wrap ##
            echo '</div>';
            
        }
        
        
        /**
         * Connect user to wp-support.co API
         * 
         * @since       0.3
         * 
         * @return      mixed     HTML message on success or False if errors encountered ##
         */
        public function ajax_q_support_connect() 
        {
            
            // check nonce ##
            check_ajax_referer( 'q_support_nonce', 'nonce' );
            
            // check for "connected" - in case there is nothing to do here ##
            if ( $this->get_q_support_user_connected() ) { 
                
                $return = __( "Account already connected", $this->text_domain );
                
            }
            
            // get plugin options ##           
            $q_support_options = $this->get_q_support_options();
            
            // get user ##
            $user = $this->get_current_user();

            // check details ##
            if ( $user && is_user_logged_in() && $user->user_login && $user->user_email ) {

                // grab a shiny new key ##
                $q_support_key = $this->generate_key();
                
                // let's dump and check some things here ##
                $q_dump = array(
                        'key'       => $q_support_key
                    ,   'username'  => $user->user_login
                    ,   'email'    => $user->user_email
                );

                // store it ##
                if ( Q_DEBUG ) { update_option( "q_support_connect_wps", $q_dump ); }
                
                // http://kovshenin.com/2011/a-note-on-wp_http-and-http-requests-in-wordpress/
                $headers    = array( 
                    'Authorization' => 'Basic ' . base64_encode( "$user->user_login:$user->user_pass" ) // very basic ##
                );
                
                // wp_remote_post call to API ## 
                $post_data = array( 
                        'method'   => 'connect'
                    ,   'username' => $user->user_login
                    ,   'email'    => $user->user_email
                    ,   'key'      => $q_support_key
                );

                // do it ! ##
                $result     = wp_remote_post ( 
                                $this->q_support_api_url 
                                ,array( 
                                        'headers'   => $headers
                                    ,   'body'      => $post_data
                                    ,   'timeout'   => $this->q_support_api_timeout, // set timeout ##
                                ) 
                            );


                if ( is_wp_error( $result ) ) {

                    $error_message = $result->get_error_message();
                    $q_support_response = $error_message;
                    $return = $error_message;
                    
                    // save it all ##
                    $this->set_q_support_settings( array( "connect_response" => $q_support_response ) );

                } else {

                    // decode the results ##
                    $results = isset( $result["body"] ) ? json_decode( $result["body"] ) : wp_die( $result ) ;
                    
                    // save API results ##
                    $this->set_q_support_settings( array( 
                            "connect_response"  => isset( $results->response ) ? $results->response : '' 
                        ,   "connect_debug"     => isset( $results->debug ) ? $results->debug : '' 
                        ,   "connect_code"      => isset( $result["response"] ) ? $result["response"] : '' 
                    ));
                    
                    // if user was created - update the "connected" and move on ##
                    if ( isset( $results->status ) && $results->status === true ) {
                        
                        // last check - for returned key ##
                        if ( ! isset( $results->key ) ) {
                            
                            $return = __( 'Unknown Connection Error', $this->text_domain );
                            
                            // save it all ##
                            $this->set_q_support_settings( array( 
                                    "connect_response"  => isset( $q_support_response ) ? $return : '' 
                            ));
                    
                        } else {
                            
                            // let's dump and check some things here ##
                            $q_dump = array(
                                    'key'           => $results->key
                            );

                            // store it ##
                            if ( Q_DEBUG ) { update_option( "q_support_connect_wps_result", $q_dump ); }
                            
                            // save it all ##
                            $this->set_q_support_settings( array( 
                                    "connected"     => true
                                ,   "key"           => $results->key
                            ));
                            
                            // add capacity 'can_request_support' to user ##
                            // http://wordpress.stackexchange.com/questions/57977/how-to-assign-capabilities-to-user-not-to-user-role/60433#60433
                            $user = new WP_User( $user->ID );
                            $user->add_cap( 'can_request_support');

                            // notice ##
                            $return = $results->response.' - '.sprintf( 
                                        __( 'You can now make <a href="%s">%s</a>', $this->text_domain )
                                        ,esc_url( $this->plugin_url )
                                        ,esc_html( __("Support Requests", $this->text_domain ) ) 
                                    );
                        
                        }
                        
                    } else { // status came back false ##

                        // just send back the API response ##
                        $return = $results->response;

                    }

                }
                
            } else {

                $return = __( "Connection Error!", $this->text_domain );

            }
            
            // return it ##
            echo json_encode($return);

            // all AJAX calls must die!! ##
            die();

        }
        
        
        /**
         * Get Plugin Default Options
         * 
         * @since       0.3
         * @return      Array  Default plugin options
         */
        public function get_q_support_options_default() 
        {
            
            // define all required options ##
            return array(
                
                // q_support options page ##
                
                // "get_server" method ##
                'get_server__phpinfo' => false,
                
            );
            
       }
        
        
        /**
         * Get Saved Plugin Settings ##
         * 
         * @since       0.3
         * @return      Array   Containing WPS Plugin settings for use inside WP ##
         */
        public function get_q_support_options() 
        {
            
            // get_option "q_support_options" - default to an empty array ##
            $q_support_options = get_option( 'q_support_options', false );
            
            // nothing cooking, so grab the default options ##
            if ( false === $q_support_options ) {
                
                #wp_die(pr( $this->get_q_support_options_default() ));
                
                // create new option for plugin settings - set to autoload ##
                add_option( 'q_support_options', $this->get_q_support_options_default(), '', 'yes' );
                
                // get_option "q_support_options" - default to an empty array ##
                $q_support_options = get_option( 'q_support_options', array() );
                
            }
            
            //return array ##
            return $q_support_options;
            
        }
        
        
        /**
         * Helper function to check if User is connected to the wp-support forum
         * 
         * @since       0.3
         * @return      boolean   
         */
        public function get_q_support_user_connected() 
        {
            
            // get user settings ##
            $q_support_settings = $this->get_q_support_settings();
            
            if ( 
                    ! $q_support_settings
                    || ! isset ( $q_support_settings["connected"] )
                    || ! isset ( $q_support_settings["key"] )
                ) 
            { 
                
                return FALSE; 
                
            } else {
                
                return TRUE;
                
            }
            
        }
        
        
        /**
         * Helper function to get user q_support_key
         * 
         * @since       0.3
         * @return      Mixed   string of user key on success | boolean false on failure   
         */
        public function get_q_support_user_key() 
        {
            
            // get user settings ##
            $q_support_settings = $this->get_q_support_settings();
            
            if ( 
                    ! $q_support_settings
                    || ! isset ( $q_support_settings["connected"] )
                    || ! isset ( $q_support_settings["key"] )
                ) 
            { 
                
                return FALSE; 
                
            } else {
                
                return $q_support_settings["key"];
                
            }
            
        }
        
        
        /**
         * Get Plugin Settings
         * 
         * @since       0.3
         * @return      Mixed    Array of plugin settings OR Boolean FALSE if errors encountered 
         */
        public function get_q_support_settings() 
        {
            
            // if the settings are missing - add the defaults ##
            if ( ! get_user_meta( $this->get_current_user("ID"), 'q_support_settings' ) ) {
                
                // save default settings ##
                update_user_meta( $this->get_current_user("ID"), 'q_support_settings', $this->q_support_settings_default );
                
            }
            
            // return settings ##
            return get_user_meta( $this->get_current_user("ID"), 'q_support_settings', true );
            
        }
        
        
        /**
         * Set User Plugin Settings
         * 
         * @since       0.3
         * @return      Boolean 
         */
        public function set_q_support_settings( $settings = null ) 
        {
            
            // sanity check ##
            if ( ! $settings || ! is_array( $settings ) ) { return false; }
            
            #pr($settings);
            
            // get user settings ##
            $q_support_settings = $this->get_q_support_settings();

            // kick back if not found - which would be odd ##
            if ( ! $q_support_settings ) { return false; }
            
            // sanitize array values ##
            $clean_by_reference = function( &$val ) {
                $val = Q_Plugin::sanitize( $val );
            };
            array_walk_recursive( $settings, $clean_by_reference );

            // loop over each key##
            foreach( $settings as $key => $value ) {
                
                $q_support_settings[sanitize_text_field($key)] = $value;
                
            }
            
            // remove keys with empty values ##
            array_filter( $q_support_settings );
            
            // update user meta ##
            update_user_meta( $this->get_current_user("ID"), 'q_support_settings', $q_support_settings );
            
        }
        
        
        /**
         * Get wp-suppport.co Agents
         * 
         * @since       0.2
         * @return      Array  list of agent names via API call 
         */
        public function get_agents() 
        {
            
            // let's cache this for 1 hour ##
            $get_agents = get_transient( 'get_agents' );
            
            if ( empty( $get_agents ) ) {
                
                // default data ##
                $get_agents = array(
                    '0' => __( 'Select', $this->text_domain )
                );   
                
                // wp_remote_post call to API ## 
                $post_data = array( 
                    'method'        => 'get_agents'
                    ,'key'          => 'agents_'.str_replace( ".", "_", Q_SUPPORT_VERSION ) // versioned key ##
                );

                // Authentication ##
                $headers    = array( 
                    'Authorization' => 'Basic ' . base64_encode( $post_data['key'] ) // very basic ##
                );

                // do it ! ##
                $result     = wp_remote_post ( 
                                $this->q_support_api_url 
                                ,array( 
                                        'headers'   => $headers
                                    ,   'body'      => $post_data
                                    ,   'timeout'   => $this->q_support_api_timeout, // set timeout ##
                                ) 
                            );


                if ( is_wp_error( $result ) ) {

                    // could give some warning here ?

                } else {

                    // decode the results ##
                    $results = isset( $result["body"] ) ? json_decode( $result["body"] ) : wp_die( $result ) ;

                    // returned true ##
                    if ( isset( $results->status ) && $results->status === true && isset( $results->response ) ) {

                        add_option( 'q_support_get_agents', $results->response );

                        foreach( $results->response as $agent ) {

                            $get_agents[$agent->ID] = $agent->user_nicename; // add each agent to the array ##

                        }

                    }

                }
                
                // save transient data - 1 hour ##
                set_transient('get_agents', $get_agents, 3600 );
                
            }
            
            // kick it back ##
            return $get_agents;
            
        }
        
        
        /**
         * Get Current User Info
         * 
         * @since       0.2
         * @link        http://codex.wordpress.org/Function_Reference/get_currentuserinfo 
         * @return      Mixed   Array user details or string of single user value
         */
        public function get_current_user( $field = null ) 
        {
            
            // user is not logged in - nothing to get ##
            if ( ! is_user_logged_in() ) { return false; }
            
            // grab the current user data ##
            global $current_user;
            get_currentuserinfo();
            
            // check if a specific field was requested from the user data - if not return the entire object ##
            if ( ! $field ) { return $current_user; }
            
            // return data ##
            return $current_user->$field;
                    
        }
        
        
        /**
         * Get WordPress Settings
         * 
         * @since       0.3
         * @return      array   Containing info about wp settings ( domain, privacy, permalinks, admin email )
         */
        private static function get_wordpress() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_wordpress'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_wordpress = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
            
                // set a new array ##
                $get_wordpress = array();
                
                // current version ##
                $get_wordpress["version"] = get_bloginfo("version");

                // latest version ##
                $cur = get_preferred_from_update_core();
                if ( ! is_object( $cur ) ) {
                    $cur = new stdClass;
                }
                
                if ( isset( $cur->current ) ) {
                    $get_wordpress["version_latest"] = $cur->current;
                }

                // version compare ##
                $compare = version_compare( $get_wordpress["version"], $get_wordpress["version_latest"], '<=' );
                $get_wordpress["is_latest"] = $compare ? true : false;
                
                $get_wordpress["admin_email"] = get_option("admin_email");
                $get_wordpress["blogname"] = get_option("blogname");
                $get_wordpress["blog_charset"] = get_option("blog_charset");
                $get_wordpress["date_format"] = get_option("date_format");
                $get_wordpress["time_format"] = get_option("time_format");
                $get_wordpress["timezone_string"] = get_option("timezone_string");
                $get_wordpress["gmt_offset"] = get_option("gmt_offset");
                $get_wordpress["home"] = get_option("home");
                $get_wordpress["siteurl"] = get_option("siteurl");
                $get_wordpress["template"] = get_option("template");
                $get_wordpress["upload_path"] = get_option("upload_path");
                $get_wordpress["posts_per_page"] = get_option("posts_per_page");
                $get_wordpress["posts_per_rss"] = get_option("posts_per_rss");
                $get_wordpress["posts_per_rss"] = get_option("posts_per_rss");
                $get_wordpress["posts_per_rss"] = get_option("posts_per_rss");
                $get_wordpress["permalink_structure"] = get_option("permalink_structure");
                $get_wordpress["blog_public"] = get_option("blog_public");
                $get_wordpress["users_can_register"] = get_option("users_can_register");
                $get_wordpress["default_role"] = get_option("default_role");
                $get_wordpress["uploads_use_yearmonth_folders"] = get_option("uploads_use_yearmonth_folders");
                $get_wordpress["upload_path"] = get_option("upload_path");
                
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_wordpress, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##

            // send back some data ##
            return $get_wordpress;
            
        }
                
                
        /**
         * Get Installed Plugins
         * 
         * @since       0.3
         * @return      array   Containing info about all active plugins ( name, is_active, version, version_latest, is_latest, author_name, author_url, support_url )
         */
        private function get_plugins() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_plugins'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_plugins = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
            
                if ( ! function_exists('get_plugins') ) {
                    require_once (ABSPATH . 'wp-admin/includes/plugin.php');
                }

                if ( ! function_exists( 'plugins_api' ) ) {
                    require_once( ABSPATH . 'wp-admin/includes/plugin-install.php' );
                }

                // grab all plugins ##
                $wp_get_plugins = get_plugins();
                #echo '<pre>'; var_dump($get_plugins); echo '</pre>';

                // loop over all plugins to build array ##
                foreach( $wp_get_plugins as $key => $value ) {

                    // unique key ##
                    $count = wp_kses( $value["Title"], $this->allowed_tags_href );

                    // Plugin Title ##
                    $get_plugins[$count]['title'] = wp_kses( $value["Title"], $this->allowed_tags_href );

                    // build plugin slug ##
                    $get_plugins[$count]['slug'] = str_replace( " ", "-", strtolower($value["Title"]) );

                    // check plugin status ##
                    if ( is_plugin_active( $key ) ) {

                        $get_plugins[$count]['status'] = 'active';

                    } else {

                        $get_plugins[$count]['status'] = 'inactive';

                    }

                    // get plugin details from comments ##
                    $get_plugins[$count]['file'] = $key;
                    $get_plugins[$count]['url'] = wp_kses( $value['PluginURI'], $this->allowed_tags_href );
                    $get_plugins[$count]['version'] = wp_kses( $value['Version'], $this->allowed_tags_href );
                    $get_plugins[$count]['version_latest'] = false;
                    $get_plugins[$count]['is_latest'] = false;
                    $get_plugins[$count]['description'] = wp_kses( $value['Description'], $this->allowed_tags_desc );
                    $get_plugins[$count]['author'] = wp_kses( $value['Author'], $this->allowed_tags_href );
                    $get_plugins[$count]['author_url'] = wp_kses( $value['AuthorURI'], $this->allowed_tags_href );

                    // set the arguments to get latest info from repository via API ##
                    $args = array(
                        'slug' => $get_plugins[$count]['slug'],
                        'fields' => array(
                            'version' => true,
                            'requires' => true,
                            'tested' => true,
                            'compatibility' => true,
                            'downloaded' => true,
                            'rating' => true,
                            'num_ratings' => true,
                            'last_updated' => true
                        )
                    );

                    // Prepare our query ##
                    $call_api = plugins_api( 'plugin_information', $args );

                    // Check for Errors & Display the results ##
                    if ( is_wp_error( $call_api ) ) {

                        $get_plugins[$count]['api_error'] = $call_api->get_error_message();
                        
                    } else {

                        #$get_plugins[$count]['api_call'] = $call_api;

                        if ( ! empty( $call_api->downloaded ) ) {

                            $get_plugins[$count]['version_latest'] = $call_api->version;
                            $get_plugins[$count]['last_updated'] = $call_api->last_updated;
                            $get_plugins[$count]['downloaded'] = $call_api->downloaded;
                            $get_plugins[$count]['requires'] = $call_api->requires;
                            $get_plugins[$count]['tested'] = $call_api->tested;
                            $get_plugins[$count]['compatibility'] = $call_api->compatibility;
                            $get_plugins[$count]['rating'] = $call_api->rating;
                            $get_plugins[$count]['num_ratings'] = $call_api->num_ratings;
                            
                            // is this the latest version ##
                            $get_plugins[$count]['is_latest'] = version_compare( $get_plugins[$count]['version'], $get_plugins[$count]['version_latest'], '<=' );
                            
                        }

                    }

                }

                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_plugins, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
            
            return $get_plugins;
            
        }
        
        
        /**
         * Get Active Widgets
         * 
         * @since       0.3
         * @return      array   Containing info about active widgets
         */
        private function get_sidebars() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_sidebars'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_sidebars = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
                
                // get all sidebars ##
                $get_registered_sidebars = $this->get_registered_sidebars();
                
                if ( $get_registered_sidebars && is_array( $get_registered_sidebars ) ) {
                    
                    foreach( $get_registered_sidebars as $sidebar ) {
                        
                        $count = wp_kses( $sidebar['id'], $this->allowed_tags_href );
                        $get_sidebars[$count]["id"] = $sidebar['id'];
                        $get_sidebars[$count]["title"] = $sidebar['name'];
                        $get_sidebars[$count]["widgets"] = $this->get_widget_data_for( $sidebar["id"] );
                        
                    }
                    
                }
                
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_sidebars, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
            
            return $get_sidebars;
            
        }
        
        
        /*
         * Get Widget data for a specified sidebar ##
         * 
         * @since       0.3
         * @return      array   Containing info about active widgets
         * @src         https://gist.github.com/kingkool68/3418186
         */
        private function get_widget_data_for( $sidebar_id ) 
        {
            
            global $wp_registered_sidebars, $wp_registered_widgets;

            // Holds the final data to return
            $output = array();

            // Loop over all of the registered sidebars looking for the one with the same ID as $sidebar_id
            //$sibebar_id = false;
            foreach( $wp_registered_sidebars as $sidebar ) {
                
                if( $sidebar['id'] == $sidebar_id ) {
                    
                    // We now have the Sidebar ID, we can stop our loop and continue.
                    $sidebar_id = $sidebar['id'];
                    break;
                    
                }
                
            }

            if( !$sidebar_id ) {
                // There is no sidebar registered with the name provided.
                return $output;
            } 

            // A nested array in the format $sidebar_id => array( 'widget_id-1', 'widget_id-2' ... );
            $sidebars_widgets = wp_get_sidebars_widgets();
            $widget_ids = $sidebars_widgets[$sidebar_id];

            if( !$widget_ids ) {
                // Without proper widget_ids we can't continue. 
                return array();
            }

            // Loop over each widget_id so we can fetch the data out of the wp_options table.
            foreach( $widget_ids as $id ) {
                
                // The name of the option in the database is the name of the widget class.  
                $option_name = $wp_registered_widgets[$id]['callback'][0]->option_name;

                // Widget data is stored as an associative array. To get the right data we need to get the right key which is stored in $wp_registered_widgets
                $key = $wp_registered_widgets[$id]['params'][0]['number'];
                
                $widget_data = get_option($option_name);
                
                // Add the widget data on to the end of the output array.
                $output[$option_name] = $widget_data[$key];
                
            }

            return $output;
            
        }
        
        
        /**
         * Get Registered Sidebars
         * 
         * @since       0.3
         * @return      array   Containing id and name of each registered sidebar - or false if the global array value is empty ##
         */
        private function get_registered_sidebars() 
        {
            
            global $wp_registered_sidebars;
            
            // nothing cooking ##
            if ( empty( $wp_registered_sidebars ) ) {
                
                return false;
                
            }
            
            // loop over each found sidebar ##
            foreach ( $wp_registered_sidebars as $sidebar ) {
                
                // unique key ##
                $count = wp_kses( $sidebar['id'], $this->allowed_tags_href );
                
                $get_registered_sidebars[$count]["id"] = $sidebar['id'];
                $get_registered_sidebars[$count]["name"] = $sidebar['name'];

            }
            
            // return it ##
            return $get_registered_sidebars;
            
        }
        
        
        /**
         * Get Active Theme info
         * 
         * @since       0.3
         * @return      array   Containing info about active theme ( name, is_active, version, version_latest, is_latest, author_name, author_url, support_url )
         */
        private function get_theme() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_theme'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_theme = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
            
                // include required functionality if missing ##
                if ( ! function_exists( 'themes_api' ) ) {
                    require_once( ABSPATH . 'wp-admin/includes/theme.php' );
                }

                // get the theme info ##
                $wp_get_theme = wp_get_theme();

                // test it ##
                #$this->pr($wp_get_theme);

                // build a new array ##
                $get_theme = array();

                // slug look-up ##
                $get_theme['slug'] = $wp_get_theme->get( 'TextDomain' ) ? $wp_get_theme->get( 'TextDomain' ) : $wp_get_theme->get( 'Name' ) ;

                // build theme slug ##
                $get_theme['slug'] = str_replace( " ", "-", strtolower( $get_theme['slug'] ) );

                // grab and format the required data ##
                $get_theme["name"] = wp_kses( $wp_get_theme->get( 'Name' ), $this->allowed_tags_href );
                $get_theme["version"] = $wp_get_theme->get( 'Version' );
                $get_theme['version_latest'] = false;
                $get_theme["is_latest"] = false; // by default ##
                $get_theme["url"] = $wp_get_theme->get( 'ThemeURI' );
                #$get_theme["description"] = wp_kses( $wp_get_theme->get( 'Description' ), $this->allowed_tags_desc );
                $get_theme["author"] = wp_kses( $wp_get_theme->get( 'Author' ), $this->allowed_tags_href );
                $get_theme["author_url"] = wp_kses( $wp_get_theme->get( 'AuthorURI' ), $this->allowed_tags_href );

                // child theme ? ##
                $get_theme["template"] = wp_kses( $wp_get_theme->get( 'Template' ), $this->allowed_tags_href );
                $get_theme["is_child"] = is_child_theme();

                // translations ##
                $get_theme["text_domain"] = wp_kses( $wp_get_theme->get( 'TextDomain' ), $this->allowed_tags_href );
                $get_theme["domain_path"] = wp_kses( $wp_get_theme->get( 'DomainPath' ), $this->allowed_tags_href );

                // set the arguments to get latest info from repository via API ##
                $args = array(
                    'slug' => $get_theme['slug'],
                    'fields' => array(
                        'version' => true,
                        'requires' => true,
                        'tested' => true,
                        'compatibility' => true,
                        'downloaded' => true,
                        'rating' => true,
                        'num_ratings' => true,
                        'last_updated' => true
                    )
                );

                // Prepare our query ##
                $call_api = themes_api( 'theme_information', $args );

                // Check for Errors & Display the results ##
                if ( is_wp_error( $call_api ) ) {

                    $get_theme['api_error'] = $call_api->get_error_message();

                } else {

                    #$get_theme['api_call'] = $call_api;

                    if ( ! empty( $call_api->downloaded ) ) {

                        $get_theme['version_latest'] = isset( $call_api->version ) ? $call_api->version : '';
                        $get_theme['last_updated'] = isset( $call_api->last_updated ) ? $call_api->last_updated : '';
                        $get_theme['downloaded'] = isset( $call_api->downloaded ) ? $call_api->downloaded : '';
                        $get_theme['requires'] = isset( $call_api->requires ) ? $call_api->requires : '';
                        $get_theme['tested'] = isset( $call_api->tested ) ? $call_api->tested : '';
                        $get_theme['compatibility'] = isset( $call_api->compatibility ) ? $call_api->compatibility : '';
                        $get_theme['rating'] = isset( $call_api->rating ) ? $call_api->rating: '';
                        $get_theme['num_ratings'] = isset( $call_api->num_ratings ) ? $call_api->num_ratings : '';

                        // is this the latest version ? ##
                        $get_theme["is_latest"] = version_compare( $get_theme["version"], $get_theme['version_latest'], '<=' );

                    }

                }
                
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_theme, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
            
            // return the array ##
            return $get_theme;
            
        }
        
        
        /**
         * Get Custom Post Types
         * 
         * @since       0.3
         * @return      array   Containing info about CPT's ##
         */
        private function get_post_types() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_post_types'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_post_types = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
                
                // get all sidebars ##
                $get_post_types = array();
                
                $post_types = get_post_types( '', 'names' ); 

                foreach ( $post_types as $post_type ) {

                    // post type name ##
                    $get_post_types[$post_type] = array();
                    $get_post_types[$post_type]["name"] = $post_type;
                    
                    // get taxonomies ##
                    $get_post_types[$post_type]["taxonomies"] = get_object_taxonomies( $post_type );
                   
                }
                
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_post_types, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
            
            return $get_post_types;
            
        }
        
        
        /**
         * Get WP_Config.php Settings
         * 
         * @since       0.3
         * @return      array   Containing info about wp_config.php ( wp_debug, wp_cache )
         * @link        http://davejesch.com/wordpress/wordpress-tech/wordpress-constants/
         * @link        http://codex.wordpress.org/Editing_wp-config.php
         */
        private static function get_wp_config() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_wp_config'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_wp_config = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
            
                // set a new array ##
                $get_wp_config = array();

                $get_wp_config['wp_home'] = defined('WP_HOME') ? WP_HOME : false;
                $get_wp_config['wp_siteurl'] = defined('WP_SITEURL') ? WP_SITEURL : false;
                $get_wp_config['wplang'] = defined('WPLANG') ? WPLANG : false;
                $get_wp_config['disable_wp_cron'] = defined('DISABLE_WP_CRON') ? DISABLE_WP_CRON : false;
                $get_wp_config['wp_cache'] = defined('WP_CACHE') ? WP_CACHE : false;
                $get_wp_config['wp_debug'] = defined('WP_DEBUG') ? WP_DEBUG : false;
                $get_wp_config['wp_debug_log'] = defined('WP_DEBUG_LOG') ? WP_DEBUG_LOG : false;
                $get_wp_config['script_debug'] = defined('SCRIPT_DEBUG') ? SCRIPT_DEBUG : false;
                $get_wp_config['wp_debug_display'] = defined('WP_DEBUG_DISPLAY') ? WP_DEBUG_DISPLAY : false;
                $get_wp_config['autosave_interval'] = defined('AUTOSAVE_INTERVAL') ? AUTOSAVE_INTERVAL : false;
                $get_wp_config['wp_post_revisions'] = defined('WP_POST_REVISIONS') ? WP_POST_REVISIONS : false;

                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_wp_config, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
                
            // send back some data ##
            return $get_wp_config;
            
        }
        
        
        /**
         * Get .htaccess settings
         * 
         * @since       0.3
         * @return      array   Containing info about wp_config.php ( file_exists, is_writeable, get_file_contents )
         */
        private static function get_htaccess() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_htaccess'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_htaccess = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
            
                // set a new array ##
                $get_htaccess = array();

                // https://github.com/Yoast/wordpress-seo/blob/master/admin/pages/files.php ##
                if ( file_exists( get_home_path() . ".htaccess" ) ) {
                    $get_htaccess["exists"] = true;
                    $get_htaccess_file = get_home_path() . ".htaccess";
                    $get_htaccess["content"] = file_get_contents($get_htaccess_file);
                    if ( is_writeable( $get_htaccess_file ) ) {
                        $get_htaccess["is_writeable"] = true;
                    }
                }
            
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_htaccess, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
                
            // send back some data ##
            return $get_htaccess;
            
        }
                
        
        /**
         * Get Server Info ##
         * 
         * @since       0.3
         * @return      array   Containing sanitized server details ##
         */
        private static function get_server() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_server'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_server = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
                
                // grab globals ##
                global $q_support_options;
                
                // set a new array ##
                $get_server = array();
            
                $get_server["phpversion"] = phpversion();
                $get_server["php_uname"] = php_uname('s');
                $get_server["php_os"] = PHP_OS;
                $get_server["server_software"] = $_SERVER['SERVER_SOFTWARE'];
                
                if ( date_default_timezone_get() ) {
                    $get_server["date_default_timezone_get"] = date_default_timezone_get();
                }

                if ( ini_get('date.timezone') ) {
                    $get_server["ini_get__date.timezone"] = ini_get('date.timezone');
                }
                
                $get_server["mysql_get_server_info"] = mysql_get_server_info();
                
                // php_info is overkill - set constant / option to true to include ##
                if ( isset( $q_support_options["get_server__phpinfo"] ) && $q_support_options["get_server__phpinfo"] === true ) $get_server["phpinfo"] = phpinfo();
                
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_server, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
                
            // send back some data ##
            return $get_server;
            
        }
        
        
        /**
         * Get Client Info ##
         * 
         * @since       0.3
         * @return      array   Containing sanitized client details ##
         */
        private static function get_client() 
        {
            
            // Generate a cache key that will hold the response for this request: 
            $transient_key = 'q_support_get_client'; 

            // Check transient. If it's there - use that, if not re fetch the theme  
            if ( false === ( $get_client = get_transient( $transient_key ) ) || Q_SUPPORT_CACHE === false ) {  
            
                // set a new array ##
                $get_client = array();
                
                // all known options ##
                // @src     http://stackoverflow.com/questions/1895727/how-can-i-detect-the-browser-with-php-or-javascript
                $browsers = array(
                    'firefox', 'msie', 'opera', 'chrome', 'safari', 'mozilla', 'seamonkey', 'konqueror', 'netscape',
                    'gecko', 'navigator', 'mosaic', 'lynx', 'amaya', 'omniweb', 'avant', 'camino', 'flock', 'aol'
                );

                if ( isset( $_SERVER['HTTP_USER_AGENT'] ) ) {
                    
                    // browser user-agent ##
                    $get_client["http_user_agent"] = $_SERVER['HTTP_USER_AGENT'];
                    
                    foreach( $browsers as $_browser ) {
                        
                        if ( preg_match( "/($_browser)[\/ ]?([0-9.]*)/", strtolower($get_client["http_user_agent"]), $match ) ) {

                            $get_client["name"] = ucfirst($match[1]);
                            $get_client["version"] = $match[2];
                            @list($get_client['majorver'], $get_client['minorver'], $get_client['build']) = explode('.', $get_client['version']);
                            break;

                        }
                          
                    }
                    
                }
                
                // Set transient for next time... keep it for 24 hours should be good  
                if ( Q_SUPPORT_CACHE === true ) set_transient( $transient_key, $get_client, Q_SUPPORT_CACHE_EXPIRE );  
            
            } // cache ##
                
            // send back some data ##
            return $get_client;
            
        }
        
        
        /*
         * Get the current URL ##
         * 
         * @since 0.3
         * @return      string  url of current page ##
         */
        public static function get_current_url()
        {
            
            // protocol ##
            $protocol = isset( $_SERVER['HTTPS'] ) ? 'https://' : 'http://';
            
            // grab the domain ##
            $domain = $_SERVER['HTTP_HOST'];
            
            // find out the path to the current file:
            $path = $_SERVER['SCRIPT_NAME'];
            
            // find out the QueryString:
            $queryString = $_SERVER['QUERY_STRING'] ? "?".$_SERVER['QUERY_STRING'] : '';

            // string it all together ##
            $url = $protocol . $domain . $path . $queryString;

            // send it back ##
            return $url;
            
        }
        
        
        /**
         * Create a compressed zip file of DUMP data 
         * 
         * @since       0.5.5
         * @return      Mixed
         */
        private function create_dump( $file_name ) 
        {
            
            // do we have a file name ? ##
            if ( ! $file_name ) { 
                
                // log error ##
                update_option( 'q_support_create_dump', 'Missing File Name' );
                
                // kick back ##
                return false; 
                
            }
            
            // grab the user settings ##
            $q_support_settings = $this->get_q_support_settings();
            
            // nothing cooking ##
            if ( ! $q_support_settings ) {
                
                // log error ##
                update_option( 'q_support_create_dump', 'Missing Settings' );
                
                // kick back ##
                return false;
                
            }
            
            // clean up the debug option ##
            delete_option( 'q_support_create_dump' );
            
            // list the settings that we need to check ##
            $settings_to_check = array(
                    'wordpress'
                ,   'server'
                ,   'client'          
            );
            
            // overwritting is cool ##
            $overwrite = true;
            
            // where are we going to save this file ##
            $path = plugin_dir_path( __FILE__ ) . '/dump/';
            
            // put it all together to get the full file path ##
            $zip_file = $path.$file_name.'.zip';
            
            // if the zip file already exists and overwrite is false, return false ##
            if( file_exists( $zip_file ) && ! $overwrite ) { return false; }
            
            // create a file of the WP install data - based on the user settings ##
            $dump = array();

            // loop over each settings area, and add some data ##
            foreach( $settings_to_check as $setting ) {
                
                // should we include this ? ##
                if ( isset( $q_support_settings[$setting] ) && $q_support_settings[$setting] == 'yes' ) {
                    
                    switch( $setting ){
                        
                        case('wordpress');
                           
                            $dump['wordpress'] = array(
                                    'htaccess'      => $this->get_htaccess()
                                ,   'wp_config'     => $this->get_wp_config()
                                ,   'wordpress'     => $this->get_wordpress()
                                ,   'post_types'    => $this->get_post_types()
                                ,   'theme'         => $this->get_theme()
                                ,   'plugins'       => $this->get_plugins()
                                ,   'sidebars'      => $this->get_sidebars()
                            );
                            
                            break;
                        
                        case('server');
                           
                            $dump['server'] = $this->get_server();
                            
                            break;
                        
                        case('client');
                           
                            $dump['client'] = $this->get_client();
                            
                            break;
                        
                    }
                    
                }
                
            }
            
            // create the archive ##
            $zip = new ZipArchive();
            if( $zip->open( $zip_file, $overwrite ? ZIPARCHIVE::OVERWRITE : ZIPARCHIVE::CREATE) !== true ) {
                
                // log error ##
                update_option( 'q_support_create_dump', 'Error Opening Zip' );
                
                // kick back ##
                return false;

            }

            // create a file inside the zip with the contents of dump ##
            $zip_addFromString = $zip->addFromString( 'dump.txt', var_export( $dump, true ) );

            // debug ##
            if ( defined( 'Q_DEBUG' ) && Q_DEBUG === true ) {

                if ( $zip_addFromString ) {

                    $zip_debug = "Zip Good: status {$zip->status}";

                } else {

                    $zip_debug = "Zip Error: status {$zip->status}";

                }

                update_option( 'q_support_create_dump', $zip_debug );

            }

            // close the zip -- done! ##
            $zip->close();

            // check to make sure the file exists ##
            if ( file_exists( $zip_file ) ) {

                return $file_name.'.zip'; // kick back the filename ##

            } else {

                return false;

            }
            
        }
        
        
        /**
         * Delete Cached Options ##
         * 
         * @since       0.3
         * @src         https://github.com/wp-plugins/delete-expired-transients/blob/master/delete-expired-transients.php
         * @return      void
         */
        protected static function delete_cache() 
        {

            global $wpdb;

            // delete all wps transients
            $sql = "
                delete from {$wpdb->options}
                where option_name like '\_transient\_q\_support\_%'
                or option_name like '\_transient\_timeout\_q\_support\_%'
            ";
                
            $wpdb->query($sql);
            
        }
        
        
        /**
         * 
         * @param       string      $dirPath
         * @throws      InvalidArgumentException
         * @link        http://stackoverflow.com/questions/3349753/delete-directory-with-files-in-it
         */
        public static function delete_directory( $dirPath ) 
        {
            
            if (! is_dir($dirPath)) {
                throw new InvalidArgumentException("$dirPath must be a directory");
            }
            
            if (substr($dirPath, strlen($dirPath) - 1, 1) != '/') {
                $dirPath .= '/';
            }
            
            $files = glob($dirPath . '*', GLOB_MARK);
            
            foreach ($files as $file) {
                if (is_dir($file)) {
                    self::delete_directory( $file );
                } else {
                    unlink($file);
                }
            }
            
            rmdir($dirPath);
            
        }
        
        
        /**
         * Clean up before a post is deleted
         * Using "before_delete_post" allows us to still access the post_metadata before that is deleted
         * 
         * @global      string        $post_type
         * @param       integer       $postid
         * @return      void
         */
        public function before_delete_post( $postid )
        {
            
            #wp_die(get_post_type( $postid ));
            
            // Check the post type ##
            if ( get_post_type( $postid ) != 'q_support' ) return;

            // get the dump url ##
            $dump = get_post_meta( $postid, 'q_support_dump', true );
            
            // key found, so let's try and find the matching dump file ##
            if ( $dump ) {
                
                // parse the URL ##
                $parse_url = parse_url( $dump );
                
                // preg_match to get our script name ##
                preg_match( '~([^/]+)\..+$~', $parse_url['path'], $script_name );
                
                // path to zipped dump file ##
                $path = plugin_dir_path( __FILE__ )."dump/$script_name[0]";
                
                // unlink it ##
                @unlink( $path );
                
                // debug ##
                if ( Q_PLUGIN_DEBUG ) update_option( "q_support_before_delete_post_{$postid}", $path );
                
            }
            
        }
        
        
        /**
         * Generate Key
         * 
         * @since       0.7
         * @return      string      Key
         */
        private function generate_key( $length = null ) 
        {
            
            // harder, but we'll need our own sanitize_key function ##
            $chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
            
            // declare our return variable ##
            $key = '';
            
            // loop over the character set a randomly defined number of times ##
            for( $i = 0; $i < ( isset( $length ) ? $length : mt_rand( 15, 50 ) ); $i++ ) {
            
                $key .= $chars[ mt_rand( 0, strlen( $chars )-1) ];
                
            }
            
            // kick it back ##
            return sanitize_file_name( $key );
            
        }
        
        
        /** 
         * Time Diff method to filter the_time function
         * 
         * @since           0.8.3
         * @link            http://www.jasonbobich.com/wordpress/a-better-way-to-add-time-ago-to-your-wordpress-theme/
         * @return          string
         */
        public function human_time_diff( $date ) 
        {

            if ( ! $date ) { return false; }
            
            #$date = get_post_time('G', true, $post);

            // Array of time period chunks
            $chunks = array(
                array( 60 * 60 * 24 * 365 , __( 'year', 'q-textdomain' ), __( 'years', 'q-textdomain' ) ),
                array( 60 * 60 * 24 * 30 , __( 'month', 'q-textdomain' ), __( 'months', 'q-textdomain' ) ),
                array( 60 * 60 * 24 * 7, __( 'week', 'q-textdomain' ), __( 'weeks', 'q-textdomain' ) ),
                array( 60 * 60 * 24 , __( 'day', 'q-textdomain' ), __( 'days', 'q-textdomain' ) ),
                array( 60 * 60 , __( 'hour', 'q-textdomain' ), __( 'hours', 'q-textdomain' ) ),
                array( 60 , __( 'minute', 'q-textdomain' ), __( 'minutes', 'q-textdomain' ) ),
                array( 1, __( 'second', 'q-textdomain' ), __( 'seconds', 'q-textdomain' ) )
            );

            if ( $date && !is_numeric( $date ) ) {
                #pr($date);
                $time_chunks = explode( ':', str_replace( ' ', ':', $date ) );
                #echo 'TIME: '.pr($time_chunks);
                $date_chunks = explode( '-', str_replace( ' ', '-', $date ) );
                #echo 'DATE: '.pr($date_chunks);
                $date = gmmktime( (int)$time_chunks[1], (int)$time_chunks[2], (int)$time_chunks[3], (int)$date_chunks[1], (int)$date_chunks[2], (int)$date_chunks[0] );
            }

            $current_time = current_time( 'mysql', $gmt = 0 );
            $newer_date = strtotime( $current_time );

            // Difference in seconds
            $since = $newer_date - $date;

            // Something went wrong with date calculation and we ended up with a negative date.
            if ( 0 > $since )
                return __( 'sometime', 'q-textdomain' );

            /**
             * We only want to output one chunks of time here, eg:
             * x years
             * xx months
             * so there's only one bit of calculation below:
             */

            //Step one: the first chunk
            for ( $i = 0, $j = count($chunks); $i < $j; $i++) {
                $seconds = $chunks[$i][0];

                // Finding the biggest chunk (if the chunk fits, break)
                if ( ( $count = floor($since / $seconds) ) != 0 )
                    break;
            }

            // Set output var
            $output = ( 1 == $count ) ? '1 '. $chunks[$i][1] : $count . ' ' . $chunks[$i][2];


            if ( !(int)trim($output) ){
                $output = '0 ' . __( 'seconds', 'q-textdomain' );
            }

            $output .= __(' ago', 'q-textdomain');

            return $output;
            
        }
            
        
    }

    // let's play :) ##
    new Q_Support;
    
// class_exists check ##
}